/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#ifndef RPARRAY_H
#define RPARRAY_H

#include "rpTypes.h"

//! Sinks an element into the heap.
template<class T>
inline void heapsink(T*rpArray, i32 element, i32 max)
{
	while ((element<<1) < max)	// there is a left child
	{
		i32 j = (element<<1);
	
		if (j+1 < max && rpArray[j] < rpArray[j+1])
			j = j+1;							// take right child

		if (rpArray[element] < rpArray[j])
		{
			T t = rpArray[j];						// swap elements
			rpArray[j] = rpArray[element];
			rpArray[element] = t;
			element = j;
		}
		else
			return;
	}
}


//! Sorts an rpArray with size 'size' using heapsort.
template<class T>
inline void heapsort(T* rpArray, i32 size)
{
	// for heapsink we pretent this is not c++, where
	// arrays start with index 0. So we decrease the rpArray pointer,
	// the maximum always +2 and the element always +1

	T* virtualArray = rpArray - 1;
	i32 virtualSize = size + 2;
	i32 i;

	// build heap

	for (i=((size-1)/2); i>=0; --i)	
		heapsink(virtualArray, i+1, virtualSize-1);

	// sort rpArray

	for (i=size-1; i>=0; --i)	
	{
		T t = rpArray[0];
		rpArray[0] = rpArray[i];
		rpArray[i] = t;
		heapsink(virtualArray, 1, i + 1);
	}
}


template <class T>
class rpArray
{

public:

	rpArray()
		: data(0), used(0), allocated(0),
			free_when_destroyed(true), is_sorted(true)
	{
	}

	//! Constructs a array and allocates an initial chunk of memory.
	//! \param start_count: Amount of elements to allocate.
	rpArray(ui32 start_count)
		: data(0), used(0), allocated(0),
			free_when_destroyed(true),	is_sorted(true)
	{
		reallocate(start_count);
	}


	//! Copy constructor
	rpArray(const rpArray<T>& other)
		: data(0)
	{
		*this = other;
	}



	//! Destructor. Frees allocated memory, if set_free_when_destroyed
	//! was not set to false by the user before.
	~rpArray()
	{
		if (free_when_destroyed)
			delete [] data;
	}



	//! Reallocates the array, make it bigger or smaller.
	//! \param new_size: New size of rpArray.
	void reallocate(ui32 new_size)
	{
		T* old_data = data;

		data = new T[new_size];
		allocated = new_size;
		
		i32 end = used < new_size ? used : new_size;
		for (i32 i=0; i<end; ++i)
			data[i] = old_data[i];

		if (allocated < used)
			used = allocated;
		
		delete [] old_data;
	}



	//! Adds an element at back of array. If the array is to small to 
	//! add this new element, the array is made bigger.
	//! \param element: Element to add at the back of the array.
	void push_back(const T& element)
	{
		if (used + 1 > allocated)
			reallocate(used * 2 +1);

		data[used++] = element;
		is_sorted = false;
	}



	//! Clears the array and deletes all allocated memory.
	void clear()
	{
		delete [] data;
		data = 0;
		used = 0;
		allocated = 0;
		is_sorted = true;
	}



	//! Sets pointer to new array, using this as new workspace.
	//! \param newPointer: Pointer to new rpArray of elements.
	//! \param size: Size of the new array.
	void set_pointer(T* newPointer, ui32 size)
	{
		delete [] data;
		data = newPointer;
		allocated = size;
		used = size;
		is_sorted = false;
	}



	//! Sets if the array should delete the memory it used.
	//! \param f: If true, the array frees the allocated memory in its
	//! destructor, otherwise not. The default is true.
	void set_free_when_destroyed(bool f)
	{
		free_when_destroyed = f;
	}



	//! Sets the size of the array.
	//! \param usedNow: Amount of elements now used.
	void set_used(ui32 usedNow)
	{
		if (allocated < usedNow)
			reallocate(usedNow);

		used = usedNow;
	}



	//! Assignement operator
	void operator=(const rpArray<T>& other)
	{
		if (data)
			delete [] data;

		//if (allocated < other.allocated)
		if (other.allocated == 0)
			data = 0;
		else
			data = new T[other.allocated];

		used = other.used;
		free_when_destroyed = other.free_when_destroyed;
		is_sorted = other.is_sorted;
		allocated = other.allocated;

		for (ui32 i=0; i<other.used; ++i)
			data[i] = other.data[i];
	}


	//! Direct access operator
	T& operator [](ui32 index)
	{
		#ifdef _DEBUG
		if (index>=used)
			_asm int 3 // access violation
		#endif

		return data[index];
	}



	//! Direct access operator
	const T& operator [](ui32 index) const
	{
		#ifdef _DEBUG
		if (index>=used)
			_asm int 3 // access violation
		#endif

		return data[index];
	}

    //! Gets last frame
	const T& getLast() const
	{
		#ifdef _DEBUG
		if (!used)
			_asm int 3 // access violation
		#endif

		return data[used-1];
	}

    //! Gets last frame
	T& getLast()
	{
		#ifdef _DEBUG
		if (!used)
			_asm int 3 // access violation
		#endif

		return data[used-1];
	}
    

	//! Returns a pointer to the array.
	//! \return Pointer to the array.
	T* pointer()
	{
		return data;
	}



	//! Returns a const pointer to the array.
	//! \return Pointer to the array.
	const T* const_pointer() const
	{
		return data;
	}



	//! Returns size of used array.
	//! \return Size of elements in the array.
	ui32 size() const
	{
		return used;
	}



	//! Returns amount memory allocated.
	//! \return Returns amount of memory allocated. The amount of bytes
	//! allocated would  be allocated_size() * sizeof(ElementsUsed);
	ui32 allocated_size() const
	{
		return allocated;
	}



	//! Returns true if array is empty
	//! \return True if the array is empty, false if not.
	bool empty() const
	{
		return used == 0;
	}



	//! Sorts the array using heapsort. There is no additional memory waste and
	//! the algorithm performs (O) n log n in worst case.
	void sort()
	{
		if (is_sorted || used<2)
			return;

		heapsort(data, used);
		is_sorted = true;
	}



	//! Performs a binary search for an element, returns -1 if not found.
	//! The array will be sorted before the binary search if it is not
	//! already sorted.
	//! \param element: Element to search for.
	//! \return Returns position of the searched element if it was found,
	//! otherwise -1 is returned.
	i32 binary_search(const T& element)
	{
		return binary_search(element, 0, used-1);
	}



	//! Performs a binary search for an element, returns -1 if not found.
	//! The array will be sorted before the binary search if it is not
	//! already sorted.
	//! \param element: Element to search for.
	//! \param left: First left index
	//! \param right: Last right index.
	//! \return Returns position of the searched element if it was found,
	//! otherwise -1 is returned.
	i32 binary_search(const T& element, i32 left, i32 right)
	{
		if (!used)
			return -1;

		sort();

		i32 m;

		do
		{
			m = (left+right)>>1;

			if (element < data[m])
				right = m - 1;
			else
				left = m + 1;

		} while((element < data[m] || data[m] < element) && left<=right);

		
		if (!(element < data[m]) && !(data[m] < element))
			return m;

		return -1;
	}


	//! Finds an element in linear time, which is very slow. Use
	//! binary_search for faster finding. Only works if =operator is implemented.
	//! \param element: Element to search for.
	//! \return Returns position of the searched element if it was found,
	//! otherwise -1 is returned.
	i32 linear_search(T& element)
	{
		for (ui32 i=0; i<used; ++i)
			if (!(element < data[i]) && !(data[i] < element))
				return (i32)i;

		return -1;
	}


	//! Finds an element in linear time, which is very slow. Use
	//! binary_search for faster finding. Only works if =operator is implemented.
	//! \param element: Element to search for.
	//! \return Returns position of the searched element if it was found,
	//! otherwise -1 is returned.
	i32 linear_reverse_search(T& element)
	{
		for (i32 i=used-1; i>=0; --i)
			if (data[i] == element)
				return (i32)i;

		return -1;
	}



	//! Erases an element from the array. May be slow, because all elements 
	//! following after the erased element have to be copied.
	//! \param index: Index of element to be erased.
	void erase(ui32 index)
	{
		#ifdef _DEBUG
		if (index>=used || index<0)
			_asm int 3 // access violation
		#endif

		for (ui32 i=index+1; i<used; ++i)
			data[i-1] = data[i];

		--used;
	}


	//! Erases some elements from the array. may be slow, because all elements 
	//! following after the erased element have to be copied.
	//! \param index: Index of the first element to be erased.
	//! \param count: Amount of elements to be erased.
	void erase(ui32 index, i32 count)
	{
		#ifdef _DEBUG
		if (index>=used || index<0 || count<1 || index+count>used)
			_asm int 3 // access violation
		#endif

		for (ui32 i=index+count; i<used; ++i)
			data[i-count] = data[i];

		used-= count;
	}


	//! Sets if the array is sorted
	void set_sorted(bool _is_sorted)
	{
		is_sorted = _is_sorted;
	}

			
	private:

		T* data;
		ui32 allocated;
		ui32 used;
		bool free_when_destroyed;
		bool is_sorted;
};



#endif