/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/



#ifndef RPCOLOR_H
#define RPCOLOR_H

#include "rpTypes.h"

//! Creates a 16 bit A1R5G5B5 color 
	inline i16 RGB16(i32 r, i32 g, i32 b)
	{
		return (((r>>3) & 0x1F)<<10) | (((g>>3) & 0x1F)<<5) | ((b>>3) & 0x1F);
	}

	//! Creates a 16 bit A1R5G5B5 color 
	inline i16 RGBA16(i32 r, i32 g, i32 b, i32 a)
	{
		return  (((a>>7) & 0x1)<<15) |
			    (((r>>3) & 0x1F)<<10) |
				(((g>>3) & 0x1F)<<5) |
				 ((b>>3) & 0x1F);
	}
	
	//! Converts a 32 bit (X8R8G8B8) color to a 16 A1R5G5B5 color
	inline i16 X8R8G8B8toA1R5G5B5(i32 color)
	{
		return RGB16(color>>16, color>>8, color);
	}

	//! Returns the red component from A1R5G5B5 color
	inline i32 getAlpha(i16 color)
	{
		return ((color >> 11)&0x1);
	}

	//! Returns the red component from A1R5G5B5 color. 
	//! Shift left by 3 to get 8 bit value.
	inline i32 getRed(i16 color)
	{
		return ((color >> 10)&0x1F);
	}


	//! Returns the green component from A1R5G5B5 color
	//! Shift left by 3 to get 8 bit value.
	inline i32 getGreen(i16 color)
	{
		return ((color >> 5)&0x1F);
	}


	//! Returns the blue component from A1R5G5B5 color
	//! Shift left by 3 to get 8 bit value.
	inline i32 getBlue(i16 color)
	{
		return ((color)&0x1F);
	}

	//! Returns the luminance from a 16 bit A1R5G5B5 color 
	inline i32 getLuminance(i16 color)
	{
		return ((getRed(color)<<3) + (getGreen(color)<<3) + (getBlue(color)<<3)) / 3;
	}


	//! Returns A8R8G8B8 Color from A1R5G5B5 color
	inline i32 A1R5G5B5toA8R8G8B8(i16 color)
	{
		return	(((color >> 15)&0x1)<<31) |	(((color >> 10)&0x1F)<<19) |
				(((color >> 5)&0x1F)<<11) |	(color&0x1F)<<3;
	}

	//! Returns A8R8G8B8 Color from R5G6B5 color
	inline i32 R5G6B5toA8R8G8B8(i16 color)
	{
		return	0xFF000000 & ((((color >> 11)&0x1F)<<19) |
				(((color >> 5)&0x3F)<<11) |	(color&0x1F)<<3);
	}


//! Class representing a 32 bit ARGB color.
/** The color values for red, green, blue
and alpha are stored in a single i32. So all four values may be between 0 and 255.
This class is used by most parts of the Irrlicht Engine
to specify a color. An other way is using the class Colorf, which
stores the color values in 4 floats.
*/
class rpColor
{
public:

	//! Constructor of the Color. Does nothing. The color value is not
	//! initialized to save time.
	inline rpColor() {}

	//! Constructs the color from 4 values representing the alpha, red, green and 
	//! blue components of the color. Must be values between 0 and 255.
	inline rpColor (i32 a, i32 r, i32 g, i32 b)
		: color(((a & 0xff)<<24) | ((r & 0xff)<<16) | ((g & 0xff)<<8) | (b & 0xff))	{}

	//! Constructs the color from a 32 bit value. Could be another color.
	inline rpColor(i32 clr)
		: color(clr) {}

	//! Returns the alpha component of the color. The alpha component
	//! defines how transparent a color should be.
	//! 0 means not transparent, 255 means fully transparent.
	inline i32 getAlpha() const	{	return (color>>24) & 0xff;	}

	//! Returns the red component of the color. 
	//! \return Returns a value between 0 and 255, specifying how red the color is.
	//! 0 means dark, 255 means full red.
	inline i32 getRed() const	{	return (color>>16) & 0xff;	}

	//! Returns the green component of the color. 
	//! \return Returns a value between 0 and 255, specifying how green the color is.
	//! 0 means dark, 255 means full green.
	inline i32 getGreen() const	{	return (color>>8) & 0xff;	}

	//! Returns the blue component of the color. 
	//! \return Returns a value between 0 and 255, specifying how blue the color is.
	//! 0 means dark, 255 means full blue.
	inline i32 getBlue() const	{	return color & 0xff;	}

	//! Sets the alpha comonent of the Color. The alpha component
	//! defines how transparent a color should be.
	//! \param a: Has to be a value between 0 and 255. 
	//! 0 means not transparent, 255 means fully transparent.
	inline void setAlpha(i32 a) { color = ((a & 0xff)<<24) | (((color>>16)& 0xff)<<16) | ((color>>8 & 0xff)<<8) | (color & 0xff); }

	//! Sets the red comonent of the Color. 
	//! \param r: Has to be a value between 0 and 255. 
	//! 0 means dark red (=black), 255 means full red.
	inline void setRed(i32 r) { color = (((color>>24) & 0xff)<<24) | ((r & 0xff)<<16) | ((color>>8 & 0xff)<<8) | (color & 0xff); }

	//! Sets the green comonent of the Color. 
	//! \param g: Has to be a value between 0 and 255. 
	//! 0 means dark green (=black), 255 means full green.
	inline void setGreen(i32 g) { color = (((color>>24) & 0xff)<<24) | (((color>>16)& 0xff)<<16) | ((g & 0xff)<<8) | (color & 0xff); }

	//! Sets the blue comonent of the Color. 
	//! \param b: Has to be a value between 0 and 255. 
	//! 0 means dark blue (=black), 255 means full blue.
	inline void setBlue(i32 b) { color = (((color>>24) & 0xff)<<24) | (((color>>16)& 0xff)<<16) | ((color>>8 & 0xff)<<8) | (b & 0xff); }

	//! Calculates a 16 bit A1R5G5B5 value of this color.
	//! \return Returns the 16 bit A1R5G5B5 value of this color.
	inline i16 toA1R5G5B5() const { return RGBA16(color>>16, color>>8, color, color>>24); };

	//! Converts color to open gl color format.
	//! \return Returns the 32 bit openGL color value.
	inline i32 toOpenGLColor() const 
	{
		return (((color>>24) & 0xff)<<24) |
				(((color)& 0xff)<<16) |
				((color>>8 & 0xff)<<8) |
				((color>>16) & 0xff);
	};

	//! Sets all four components of the color at once.
	//! Constructs the color from 4 values representing the alpha, red, green and 
	//! blue components of the color. Must be values between 0 and 255.
	//! \param a: Alpha component of the color. 
	//! The alpha component defines how transparent a color should be.
	//! Has to be a value between 0 and 255. 
	//! 0 means not transparent, 255 means fully transparent.
	//! \param r: Sets the red comonent of the Color. 
	//! Has to be a value between 0 and 255. 
	//! 0 means dark red (=black), 255 means full red.
	//! \param g: Sets the green comonent of the Color. 
	//! Has to be a value between 0 and 255. 
	//! 0 means dark green (=black), 255 means full green.
	//! \param b: Sets the blue comonent of the Color. 
	//! Has to be a value between 0 and 255. 
	//! 0 means dark blue (=black), 255 means full blue.
	inline void set(i32 a, i32 r, i32 g, i32 b) { color = (((a & 0xff)<<24) | ((r & 0xff)<<16) | ((g & 0xff)<<8) | (b & 0xff)); }

	//! Compares the color to another color.
	//! \return Returns true if the colors are the same, and false if not.
	inline bool operator==(const rpColor& other) const { return other.color == color; }

	//! Compares the color to another color.
	//! \return Returns true if the colors are different, and false if they are the same.
	inline bool operator!=(const rpColor& other) const { return other.color != color; }

	//! Interpolates the color with a f32 value to an other color 
	//! \param other: Other color
	//! \param d: value between 0.0f and 1.0f
	//! \return Returns interpolated color.
	inline rpColor getInterpolated(rpColor other, f32 d) const
	{
		f32 inv = 1.0f - d;
		return rpColor((i32)(other.getAlpha()*inv + getAlpha()*d),
			(i32)(other.getRed()*inv + getRed()*d),
			(i32)(other.getGreen()*inv + getGreen()*d),
			(i32)(other.getBlue()*inv + getBlue()*d));
	}
    
	//! color in A8R8G8B8 Format
	i32 color; 
};


//! Class representing a color with four floats.
/**	The color values for red, green, blue
and alpha are each stored in a 32 bit floating point variable.
So all four values may be between 0.0f and 1.0f.
This class is used rarely used by the Irrlicht Engine
to specify a color. An other, faster way is using the class Color, which
stores the color values in a single 32 bit integer. 
*/
class rpColorf
{
public:

	//! Constructs a color. All values are initialised with 0.0f, resulting
	//! in a black color.
	rpColorf() : r(0.0f), g(0.0f), b(0.0f), a(0.0f) {};

	//! Constructs a color from three color values: red, green and blue.
	//! \param r: Red color component. Should be a value between 0.0f meaning
	//! very dark red (=black) and 1.0f, meaning full red.
	//! \param g: Green color component. Should be a value between 0.0f meaning
	//! very dark green (=black) and 1.0f, meaning full green.
	//! \param b: Blue color component. Should be a value between 0.0f meaning
	//! very dark blue (=black) and 1.0f, meaning full blue.
	rpColorf(f32 r, f32 g, f32 b) : r(r), g(g), b(b), a(1.0f) {};

	//! Constructs a color from four color values: alpha, red, green and blue.
	//! \param a: Alpha color component of the color.
	//! The alpha component defines how transparent a color should be.
	//! Has to be a value between 0.0f and 1.0f,  
	//! 0.0f means not transparent, 1.0f means fully transparent.
	//! \param r: Red color component. Should be a value between 0.0f meaning
	//! very dark red (=black) and 1.0f, meaning full red.
	//! \param g: Green color component. Should be a value between 0.0f meaning
	//! very dark green (=black) and 1.0f, meaning full green.
	//! \param b: Blue color component. Should be a value between 0.0f meaning
	//! very dark blue (=black) and 1.0f, meaning full blue.
	rpColorf(f32 r, f32 g, f32 b, f32 a) : r(r), g(g), b(b), a(a) {};

	//! Constructs a color from 32 bit Color.
	//! \param c: 32 bit color value from which this Colorf class is 
	//! constructed from.
	rpColorf(rpColor c) { const f32 inv = 1.0f / 255.0f; r = c.getRed() * inv; g = c.getGreen() * inv; b = c.getBlue() * inv;	a = c.getAlpha() * inv;	};

	//! Converts this color to a rpColor without floats.
	rpColor torpColor() const
	{
		return rpColor((i32)(a*255.0f), (i32)(r*255.0f), (i32)(g*255.0f), (i32)(b*255.0f));
	}

	//! red color component
	f32 r;

	//! green color component
	f32 g;

	//! blue component
	f32 b;

	//! alpha color component
	f32 a;

	//! Sets the three color comonents to new values at once.
	//! \param rr: Red color component. Should be a value between 0.0f meaning
	//! very dark red (=black) and 1.0f, meaning full red.
	//! \param gg: Green color component. Should be a value between 0.0f meaning
	//! very dark green (=black) and 1.0f, meaning full green.
	//! \param bb: Blue color component. Should be a value between 0.0f meaning
	//! very dark blue (=black) and 1.0f, meaning full blue.
	void set(f32 rr, f32 gg, f32 bb) {r = rr; g =gg; b = bb; };
};

// Copyright (C) 2002-2005 Nikolaus Gebhardt
// This file is part of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in Irrlicht.h

class rpColorConverter
{
public:

	//! converts a 4 bit palettized image into R5G5B5
	static void convert4BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch, const i32* palette);

	//! converts a 8 bit palettized image into R5G5B5
	static void convert8BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch, const i32* palette);

	//! converts a 8 bit palettized image into R5G5B5
	static void convert8BitTo16Bit(const c8* in, i16* out, i32 width, i32 height, i32 pitch, const i32* palette);

	//! converts a monochrome bitmap to A1R5G5B5 data
	static void convert1BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch);

	//! converts R8G8B8 24 bit data to A1R5G5B5 data, and flips and 
	//! mirrors the image during the process.
	static void convert24BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch);

	//! converts R8G8B8 16 bit data to A1R5G5B5 data, and flips and 
	//! mirrors the image during the process.
	static void convert16BitTo16BitFlipMirror(const i16* in, i16* out, i32 width, i32 height, i32 pitch);

	//! copies R8G8B8 24 bit data to 24 data, and flips and 
	//! mirrors the image during the process.
	static void convert24BitTo24BitFlipMirrorColorShuffle(const c8* in, c8* out, i32 width, i32 height, i32 pitch);

	//! converts R8G8B8 24 bit data to A1R5G5B5 data (used e.g for JPG to A1R5G5B5)
	//! accepts colors in different order.
	static void convert24BitTo16BitColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch);

	//! converts R8G8B8 24 bit data to A1R5G5B5 data (used e.g for JPG to A1R5G5B5)
	//! accepts colors in different order.
	static void convert24BitTo16BitFlipColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch);

	//! converts X8R8G8B8 32 bit data to A1R5G5B5 data, and flips and 
	//! accepts colors in different order.
	static void convert32BitTo16BitColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch);

	//! converts X8R8G8B8 32 bit data to A1R5G5B5 data, and flips and 
	//! mirrors the image during the process, accepts colors in different order.
	static void convert32BitTo16BitFlipMirrorColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch);

	//! copies X8R8G8B8 32 bit data, and flips and 
	//! mirrors the image during the process.
	static void convert32BitTo32BitFlipMirror(const i32* in, i32* out, i32 width, i32 height, i32 pitch);

	//! Resizes the surface to a new size and converts it at the same time
	//! to an A8R8G8B8 format, returning the pointer to the new buffer.
	//! The returned pointer has to be deleted.
	static void convert16bitToA8R8G8B8andResize(const i16* in, i32* out, i32 newWidth, i32 newHeight, i32 currentWidth, i32 currentHeight);
};




#endif
