/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#ifndef RPIMAGE_H
#define RPIMAGE_H
#include "rpbase.h"
#include "rpColor.h"

enum E_COLOR_FORMAT
{
	//! 16 bit color format used by the software driver, and thus preferred
	//! by all other irrlicht engine video drivers. There are 5 bits for every
	//! color component, and a single bit is left for alpha information.
	E_CF_A1R5G5B5 = 0,	

	//! Standard 16 bit color format.
	E_CF_R5G6B5,

	//! 24 bit color, no alpha channel, but 8 bit for red, green and blue.
	E_CF_R8G8B8,

	//! Default 32 bit color format. 8 bits are used for every component:
	//! red, green, blue and alpha.
	E_CF_A8R8G8B8,
};


class rpImage :
	public rpBase
{
public:
	
	//! constructor 
	rpImage(E_COLOR_FORMAT format, rpImage* imageToCopy);

	//! constructor 
	rpImage(E_COLOR_FORMAT format, const dimension2d<i32>& size,
		void* data);

	//! constructor
	rpImage(E_COLOR_FORMAT format, const dimension2d<i32>& size);

	//! constructor
	rpImage(rpImage* imageToCopy,
		const position2d<i32>& pos, const dimension2d<i32>& size);

	//! destructor
	virtual ~rpImage();

	//! Lock function.
	virtual void* lock();

	//! Unlock function.
	virtual void unlock();

	//! Returns width and height of image data.
	virtual const dimension2d<i32>& getDimension();

	//! Returns bits per pixel. 
	virtual i32 getBitsPerPixel();

	//! Returns bytes per pixel
	virtual i32 getBytesPerPixel();

	//! Returns image data size in bytes
	virtual i32 getImageDataSizeInBytes();

	//! Returns image data size in pixels
	virtual i32 getImageDataSizeInPixels();

	//! returns mask for red value of a pixel
	virtual i32 getRedMask();

	//! returns mask for green value of a pixel
	virtual i32 getGreenMask();

	//! returns mask for blue value of a pixel
	virtual i32 getBlueMask();

	//! returns mask for alpha value of a pixel
	virtual i32 getAlphaMask();

	//! returns a pixel
	virtual rpColor getPixel(i32 x, i32 y);

	//! returns the color format
	virtual E_COLOR_FORMAT getColorFormat();

	//! draws a rectangle
	void drawRectangle(const rect<i32>& rect, rpColor color);

	//! draws a rectangle
	void drawRectangle(i32 x, i32 y, i32 x2, i32 y2, rpColor color);

	//! copies this surface into another
	void copyTo(rpImage* target, i32 x, i32 y);

	//! copies this surface into another
	void copyTo(rpImage* target, const position2d<i32>& pos);

	//! copies this surface into another
	void copyTo(rpImage* target, const position2d<i32>& pos, const rect<i32>& sourceRect, const rect<i32>* clipRect=0);

	//! copies this surface into another, using the alpha mask
	void copyToWithAlpha(rpImage* target, const position2d<i32>& pos, const rect<i32>& sourceRect);

	//! copies this surface into another, using the alpha mask, an cliprect and a color to add with
	void copyToWithAlpha(rpImage* target, const position2d<i32>& pos, const rect<i32>& sourceRect, i16 color, const rect<i32>* clipRect = 0);

	//! copies this surface into another, scaling it to fit it.
	void copyToScaling(rpImage* target);

	//! draws a line from to
	void drawLine(const position2d<i32>& from, const position2d<i32>& to, i16 color);
	
	//! resizes the surface to a new size
	void resizeTo(const dimension2d<i32>& size);

	//! fills the surface with black or white
	void fill(i16 color);
	
	
private:

	//! assumes format and size has been set and creates the rest
	void initData();

	//! clips a x coordinate into the screen
	inline void clipX(i32 &x);

	//! clips a y coordinate into the screen
	inline void clipY(i32 &y);

	//! exchanges two ints
	inline void exchange(i32& a, i32& b);

	//! clips coordinates into the rect. returns false if the rects is completely outside
	//! and does not need to be drawn.
	inline bool clipRect(const i32 targetWidth, const i32 targetHeight, i32& targetX, i32& targetY, i32& xInSource, i32& yInSource, i32& sourceWidth, i32& soureHeight);

	//! sets a pixel very fast and inline
	inline void setPixelFast(i32 x, i32 y, i16 color);

	i32 getBitsPerPixelFromFormat();

	void* Data;
	dimension2d<i32> Size;
	i32 BitsPerPixel;
	i32 BytesPerPixel;
	E_COLOR_FORMAT Format;

	i32 RedMask;
	i32 GreenMask;
	i32 BlueMask;
	i32 AlphaMask;

};


#endif