/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#ifndef RPVECTOR_H
#define RPVECTOR_H

#include <math.h>
#include "rpMath.h"
#include "rpTypes.h"

//! 2d vector template class with lots of operators and methods.
template <class T>
class rpVector2D
{
public:

	// member variables
	T X, Y;

	rpVector2D(): X(0), Y(0) {};
	rpVector2D(T nx, T ny) : X(nx), Y(ny) {};
	rpVector2D(const rpVector2D<T>& other)	:X(other.X), Y(other.Y) {};

	// operators

	rpVector2D<T> operator-() const { return rpVector2D<T>(-X, -Y);   }

	rpVector2D<T>& operator=(const rpVector2D<T>& other)	{ X = other.X; Y = other.Y; return *this; }

	rpVector2D<T> operator+(const rpVector2D<T>& other) const { return rpVector2D<T>(X + other.X, Y + other.Y);	}
	rpVector2D<T>& operator+=(const rpVector2D<T>& other)	{ X+=other.X; Y+=other.Y; return *this; }

	rpVector2D<T> operator-(const rpVector2D<T>& other) const { return rpVector2D<T>(X - other.X, Y - other.Y);	}
	rpVector2D<T>& operator-=(const rpVector2D<T>& other)	{ X-=other.X; Y-=other.Y; return *this; }

	rpVector2D<T> operator*(const rpVector2D<T>& other) const { return rpVector2D<T>(X * other.X, Y * other.Y);	}
	rpVector2D<T>& operator*=(const rpVector2D<T>& other)	{ X*=other.X; Y*=other.Y; return *this; }
	rpVector2D<T> operator*(const T v) const { return rpVector2D<T>(X * v, Y * v);	}
	rpVector2D<T>& operator*=(const T v) { X*=v; Y*=v; return *this; }

	rpVector2D<T> operator/(const rpVector2D<T>& other) const { return rpVector2D<T>(X / other.X, Y / other.Y);	}
	rpVector2D<T>& operator/=(const rpVector2D<T>& other)	{ X/=other.X; Y/=other.Y; return *this; }
	rpVector2D<T> operator/(const T v) const { return rpVector2D<T>(X / v, Y / v);	}
	rpVector2D<T>& operator/=(const T v) { X/=v; Y/=v; return *this; }

	bool operator==(const rpVector2D<T>& other) const { return other.X==X && other.Y==Y; }
	bool operator!=(const rpVector2D<T>& other) const { return other.X!=X || other.Y!=Y; }

	// functions

	void set(const T& nx, const T& ny) {X=nx; Y=ny; }
	void set(const rpVector2D<T>& p) { X=p.X; Y=p.Y;}

	//! Returns the length of the vector
	//! \return Returns the length of the vector.
	f64 getLength() const { return sqrt(X*X + Y*Y); }

	//! Returns the dot product of this vector with an other.
	T dotProduct(const rpVector2D<T>& other) const
	{
		return X*other.X + Y*other.Y;
	}

	//! Returns distance from an other point. Here, the vector is interpreted as
	//! point in 2 dimensional space.
	f64 getDistanceFrom(const rpVector2D<T>& other) const
	{
		f64 vx = X - other.X; f64 vy = Y - other.Y;
		return sqrt(vx*vx + vy*vy);
	}

	//! rotates the point around a center by an amount of degrees.
	void rotateBy(f64 degrees, const rpVector2D<T>& center)
	{
		degrees *= GRAD_PI2;
		T cs = (T)cos(degrees);
		T sn = (T)sin(degrees);

		X -= center.X;
		Y -= center.Y;

		set(X*cs - Y*sn, X*sn + Y*cs);

		X += center.X;
		Y += center.Y;
	}

	//! normalizes the vector.
	rpVector2D<T>& normalize()
	{
		T l = (T)getLength();
		if (l == 0)
			return *this;

		l = (T)1.0 / l;
		X *= l;
		Y *= l;
		return *this;
	}

	//! Calculates the angle of this vector in grad.
	//! \return Returns a value between 0 and 360.
	inline f64 getAngle() const
	{
		if (Y == 0.0)  // corrected thanks to a suggestion by Jox
			return X < 0.0 ? 180.0 : 0.0; 
		else if (X == 0.0) 
			return Y < 0.0 ? 90.0 : 270.0;

		f64 tmp = Y / sqrt(X*X + Y*Y);
		tmp = atan(sqrt(1 - tmp*tmp) / tmp) * GRAD_PI;

		if (X>0.0 && Y>0.0)
			return tmp + 270;
		else
		if (X>0.0 && Y<0.0)
			return tmp + 90;
		else
		if (X<0.0 && Y<0.0)
			return 90 - tmp;
		else
		if (X<0.0 && Y>0.0)
			return 270 - tmp;

		return tmp;
	}

	//! Calculates the angle between this vector and another one in grad.
	//! \return Returns a value between 0 and 90.
	inline f64 getAngleWith(const rpVector2D<T>& b) const
	{
		f64 tmp = X*b.X + Y*b.Y;

		if (tmp == 0.0)
			return 90.0;

		tmp = tmp / sqrt((X*X + Y*Y) * (b.X*b.X + b.Y*b.Y));
		if (tmp < 0.0) tmp = -tmp;

		return atan(sqrt(1 - tmp*tmp) / tmp) * GRAD_PI;
	}

	
};




class rpVector3D {

public:
    float x, y, z;

    rpVector3D() {
        x = (float) 0.0;
        y = (float) 0.0;
        z = (float) 0.0;
    };
    
    rpVector3D(float _x, float _y, float _z) {
        x = _x;
        y = _y;
        z = _z;
    };
    
    void  Set(float _x, float _y, float _z) {
        x = _x;
        y = _y;
        z = _z;
    };
    
    float Len() {
        return rp_sqrt( x*x + y*y + z*z );
    };
    
    void  Norm();
	void  Rotate( rpVector3D *axis, float angle );
	void  RotateX( float angle );
	void  RotateY( float angle );
	void  RotateZ( float angle );
	void  Reset( void );

    // operators
    rpVector3D& operator*= ( const rpVector3D& first );
    rpVector3D& operator+= ( const rpVector3D& first );
    rpVector3D& operator-= ( const rpVector3D& first );
    rpVector3D& operator*= ( float first);
};

// methods 

inline rpVector3D  operator+  ( const rpVector3D& first, const rpVector3D& second );
inline rpVector3D  operator-  ( const rpVector3D& first, const rpVector3D& second );
inline rpVector3D  operator*  ( const rpVector3D& first, const rpVector3D& second );
inline float      operator%  ( const rpVector3D& first, const rpVector3D& second );
inline rpVector3D  operator*  ( float      first, const rpVector3D& second );
inline rpVector3D  operator*  ( const rpVector3D& first, float      second );
inline rpVector3D  operator/  ( const rpVector3D& first, float      second );



// normalize
inline void rpVector3D::Norm()
{
    float n;
    n = Len();
    if( n > 0.0f ) {
        n   = 1 / n;
        x = x * n;
        y = y * n;
        z = z * n;
        }
    else {
        x = (float) 0.0;
        y = (float) 0.0;
        z = (float) 0.0;
        }

    return;
}


inline void rp_vec_rot_round_vec( rpVector3D *vec, float *m )
{
	rpVector3D v;
	v = (*vec);

	vec->x = m[0] * v.x + m[1] * v.y + m[2] * v.z;
	vec->y = m[3] * v.x + m[4] * v.y + m[5] * v.z;
	vec->z = m[6] * v.x + m[7] * v.y + m[8] * v.z;
}

inline void rpVector3D::Rotate( rpVector3D *axis, float angle )
{
	
	float rot_m[9];
	rpVector3D axis2;
	float sa, ca;

	sa = (float) sin( angle );
	ca = (float) cos( angle );

	
	axis2 = (*axis);
	axis2.Norm();

	
	rot_m[0] = ca + (1 - ca) * axis2.x * axis2.x;
	rot_m[1] = (1 - ca) * axis2.x * axis2.y - sa * axis2.z;
	rot_m[2] = (1 - ca) * axis2.z * axis2.x + sa * axis2.y;
	rot_m[3] = (1 - ca) * axis2.x * axis2.y + sa * axis2.z;
	rot_m[4] = ca + (1 - ca) * axis2.y * axis2.y;
	rot_m[5] = (1 - ca) * axis2.y * axis2.z - sa * axis2.x;
	rot_m[6] = (1 - ca) * axis2.z * axis2.x - sa * axis2.y;
	rot_m[7] = (1 - ca) * axis2.y * axis2.z + sa * axis2.x;
	rot_m[8] = ca + (1 - ca) * axis2.z * axis2.z;

	
	rp_vec_rot_round_vec( this, rot_m );
}


inline void rpVector3D::RotateX( float a )
{
	 
	float rot_m[9];

	rot_m[0] = 1.0f; rot_m[1] = 0.0f;            rot_m[2] =  0.0f;
	rot_m[3] = 0.0f; rot_m[4] = (float) cos(a); rot_m[5] = -(float) sin(a);
	rot_m[6] = 0.0f; rot_m[7] = (float) sin(a); rot_m[8] =  (float) cos(a);

	rp_vec_rot_round_vec( this, rot_m );
}

inline void rpVector3D::RotateY( float a )
{
	float rot_m[9];

	rot_m[0] = (float) cos(a);  rot_m[1] = 0.0f; rot_m[2] = (float) sin(a);
	rot_m[3] = 0.0f;            rot_m[4] = 1.0f; rot_m[5] = 0.0f;
	rot_m[6] = -(float) sin(a); rot_m[7] = 0.0f; rot_m[8] = (float) cos(a);

	rp_vec_rot_round_vec( this, rot_m );
}

inline void rpVector3D::RotateZ( float a )
{
	float rot_m[9];

	rot_m[0] = (float) cos(a); rot_m[1] = -(float) sin(a); rot_m[2] = 0.0f;
	rot_m[3] = (float) sin(a); rot_m[4] =  (float) cos(a); rot_m[5] = 0.0f;
	rot_m[6] = 0.0f;           rot_m[7] = 0.0f;            rot_m[8] = 1.0f;

	rp_vec_rot_round_vec( this, rot_m );
}

inline void rpVector3D::Reset( void )
{
    this->x = 0.0;
    this->y = 0.0;
    this->z = 0.0;
}


// operators
inline rpVector3D operator+ ( const rpVector3D& first, const rpVector3D& second )
{
    return rpVector3D( first.x + second.x,
                      first.y + second.y,
                      first.z + second.z );
}

inline rpVector3D operator- ( const rpVector3D& first, const rpVector3D& second )
{
    return rpVector3D( first.x - second.x,
                      first.y - second.y,
                      first.z - second.z );
}

inline rpVector3D operator* ( const rpVector3D& first, const rpVector3D& second )
{
    return rpVector3D ( first.y * second.z - first.z * second.y,
                       first.z * second.x - first.x * second.z,
                       first.x * second.y - first.y * second.x );
}

inline float operator% ( const rpVector3D& first, const rpVector3D& second )
{
    return ( first.x * second.x + first.y * second.y + first.z * second.z );
}


inline rpVector3D operator* ( const rpVector3D& first, float second )
{
    return rpVector3D( first.x * second,
                      first.y * second,
                      first.z * second );
}

inline rpVector3D operator* ( float first, const rpVector3D& second )
{
    return rpVector3D( first * second.x,
                      first * second.y,
                      first * second.z );
}

inline rpVector3D operator/ ( const rpVector3D& first, float second )
{
    return rpVector3D( first.x / second,
                      first.y / second,
                      first.z / second );
}

inline rpVector3D& rpVector3D::operator+= ( const rpVector3D& second )
{
    x += second.x;
    y += second.y;
    z += second.z;

    return *this;
}

inline rpVector3D& rpVector3D::operator-= ( const rpVector3D& second )
{
    x -= second.x;
    y -= second.y;
    z -= second.z;

    return *this;
}

inline rpVector3D& rpVector3D::operator*= ( float second )
{
    x *= second;
    y *= second;
    z *= second;

    return *this;
}

inline rpVector3D& rpVector3D::operator*= ( const rpVector3D& second )
{
    (*this) = (*this) * second;

    return *this;
}



#endif
