/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#ifndef RPVIDEODRIVER_H
#define RPVIDEODRIVER_H

#include "rpbase.h"
#include "rpMatrix.h"
#include "rpVector.h"
#include "rpColor.h"
#include "rpTexture.h"


enum E_VIDEO_DRIVER {
		//opengl driver
		E_VD_OPENGL = 0,

		//never used
		E_VD_COUNT
	};

enum E_VIDEO_DRIVER_FEATURE
{
	// Is driver able to render to a surface?
	E_VDF_RENDER_TO_TARGET = 0,	

	// Is driver able to render with a bilinear filter applied?
	E_VDF_BILINEAR_FILER,			

	// Is hardeware transform and lighting supported?
	E_VDF_HARDWARE_TL,			

	// Can the driver handle mip maps?
	E_VDF_MIP_MAP,		

	// Are stencilbuffers switched on and does the device support stencil buffers?
	E_VDF_STENCIL_BUFFER,

	// Is Vertex Shader 1.1 supported?
	E_VDF_VERTEX_SHADER_1_1,

	// Is Vertex Shader 2.0 supported?
	E_VDF_VERTEX_SHADER_2_0,

	// Is Vertex Shader 3.0 supported?
	E_VDF_VERTEX_SHADER_3_0,

	// Is Pixel Shader 1.1 supported?
	E_VDF_PIXEL_SHADER_1_1,

	// Is Pixel Shader 1.2 supported?
	E_VDF_PIXEL_SHADER_1_2,

	// Is Pixel Shader 1.3 supported?
	E_VDF_PIXEL_SHADER_1_3,

	// Is Pixel Shader 1.4 supported?
	E_VDF_PIXEL_SHADER_1_4,

	// Is Pixel Shader 2.0 supported?
	E_VDF_PIXEL_SHADER_2_0,

	// Is Pixel Shader 3.0 supported?
	E_VDF_PIXEL_SHADER_3_0,

	// Are ARB vertex programs v1.0 supported?
	E_VDF_ARB_VERTEX_PROGRAM_1,

	// Are ARB fragment programs v1.0 supported?
	E_VDF_ARB_FRAGMENT_PROGRAM_1,
};

// enumeration for geometry transformation states
enum E_TRANSFORMATION_STATE
{
	// View transformation
	E_TS_VIEW = 0,
	// World transformation
	E_TS_WORLD,
	// Projection transformation
	E_TS_PROJECTION,
	// Not used
	E_TS_COUNT
};
class rpGeometryBuffer;
class rpLightNode;

class rpVideoDriver :
	public rpBase
{
public:
	
	rpVideoDriver(const dimension2d<i32>&);
	~rpVideoDriver(void);

	
	virtual bool	beginScene(bool backBuffer, bool zBuffer, rpColorf color);

	//! Presents the rendered image on the screen.
	virtual bool	endScene();

	//! Queries the features of the driver, returns true if a feature is available
	virtual bool	queryFeature(E_VIDEO_DRIVER_FEATURE feature) = 0;

	//! Sets the view, world or projection transformation. 
	virtual void	setTransform(E_TRANSFORMATION_STATE state, rpMatrix& mat) = 0;

	//! Returns the transformation set by setTransform
	virtual			rpMatrix getTransform(E_TRANSFORMATION_STATE state) = 0;

	virtual void	setFog(rpColor color=rpColor(0,255,255,255), bool linearFog=true, f32 start=50.0f, f32 end=100.0f, 
						f32 density=0.01f, bool pixelFog=false, bool rangeFog=false) = 0;

	virtual void	setTextureCreationFlag(E_TEXTURE_CREATION_FLAG flag, bool enabled) = 0;

	virtual void	setTexture(rpTexture *) = 0;


	// device dependent texture handling
	virtual rpTexture*			newTexture() = 0;

	// device dependent vertex buffers
	virtual rpGeometryBuffer*	newGeometryBuffer() = 0;

	// render device dependent geometry buffer
	virtual void renderGeometryBuffer(rpGeometryBuffer *) = 0;

	// handle lights
	virtual void	addDynamicLight(rpLightNode *) = 0;
	virtual void	removeDynamicLights() = 0;


	//performance statistics
	i32		getLastFramesCount();
	f32		getLastFramesPerSec();
	void	setStatsInterval(i32 ms) {
		msec_treshold = ms;
	}


	
protected:

	// current output viewport
	rect<i32> viewport;
	// size of the rendering window/screen
	dimension2d<i32> screenSize;
	// statistics only
	ui32 primitivesDrawn;

	// current transformation matrices presented on the driver
	rpMatrix	matrix[E_TS_COUNT];

	
	ui32 textureCreationFlags;

	// fog atributes
	bool linearFog;
	f32 fogStart;
	f32 fogEnd;
	f32 fogDensity;
	bool pixelFog;
	bool rangeFog;
	rpColor fogColor;

	// performance statistics
	i32		startup_secs;
	i32		startup_msec;
	i32		msec_duration;
	i32		msec_treshold;

	i32		frames_rendered;
	i32		last_frames_rendered;
};


#endif