/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#include "rpColor.h"

//! converts a 4 bit palettized image into R5G5B5
void rpColorConverter::convert4BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch, const i32* palette)
{
	i32 shift = 0;
	out += width*height;
	i16* oout = out;

	for (i32 y=0; y<height; ++y)
	{
		shift = 4;

		out = oout - (y*width) - width;

		for (i32 x=0; x<width; ++x)
		{
			*out = X8R8G8B8toA1R5G5B5(palette[(uc8)((*in >> shift) & 0xf)]);
			++out;

			shift -= 4;
			if (shift<0)
			{
				shift = 4;
				++in;
			}
		}

		if (shift !=4)
			++in;

		in+=pitch;
	}
}

//! converts a 8 bit palettized image into R5G5B5
void rpColorConverter::convert8BitTo16Bit(const c8* in, i16* out, i32 width, i32 height, i32 pitch, const i32* palette)
{
	i32 lineWidth = width + pitch;
	const c8* p = in;

	for (i32 y=1; y<=height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			
			*out = X8R8G8B8toA1R5G5B5(palette[(uc8)(*p)]);

			++out;
			++p;
		}

		p= in + y*lineWidth + pitch;
	}
}



//! converts a 8 bit palettized image into R5G5B5
void rpColorConverter::convert8BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch, const i32* palette)
{
	out += width * height;
	i32 lineWidth = width + pitch;
	const c8* p = in;

	for (i32 y=1; y<=height; ++y)
	{
		p= in + lineWidth*y - pitch;

		for (i32 x=0; x<width; ++x)
		{
			--out;
			--p;
			*out = X8R8G8B8toA1R5G5B5(palette[(uc8)(*p)]);
		}
	}
}


//! converts a monochrome bitmap to A1R5G5B5 data
void rpColorConverter::convert1BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch)
{
	i16* p = out + width * height;

	for (i32 y=0; y<height; ++y)
	{
		i32 shift = 7;
		out = p - y * width - width;

		for (i32 x=0; x<width; ++x)
		{
			*out = *in>>shift & 0x01 ? (i16)0xffff : (i16)0x0000;
			++out;
							
			--shift;
			if (shift<0)
			{
				shift=7;
				++in;
			}
		}

		if (shift != 7)
			++in;

		in += pitch;
	}
}

//! converts R8G8B8 16 bit data to A1R5G5B5 data, and flips and 
//! mirrors the image during the process.
void rpColorConverter::convert16BitTo16BitFlipMirror(const i16* in, i16* out, i32 width, i32 height, i32 pitch)
{
	const i16 *p = in;
	const i32 lineWidth = width + pitch;
	out += width * height;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (width-x-1);
			--out;
			*out = *p;
		}

		in += lineWidth;
	}
}


//! converts R8G8B8 24 bit data to A1R5G5B5 data, and flips and 
//! mirrors the image during the process.
void rpColorConverter::convert24BitTo16BitFlipMirror(const c8* in, i16* out, i32 width, i32 height, i32 pitch)
{
	const c8 *p = in;
	const i32 lineWidth = 3 * width + pitch;
	out += width * height;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (width-x-1)*3;
			--out;
			*out = RGB16(p[2], p[1], *p);
		}

		in += lineWidth;
	}
}



//! copies R8G8B8 24 bit data to 24 data, and flips and 
//! mirrors the image during the process.
void rpColorConverter::convert24BitTo24BitFlipMirrorColorShuffle(const c8* in, c8* out, i32 width, i32 height, i32 pitch)
{
	const c8 *p = in;
	const i32 lineWidth = 3 * width + pitch;
	out += width * height * 3;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (width-x-1)*3;
			out -= 3;
			out[0] = p[2];
			out[1] = p[1];
			out[2] = p[0];
		}

		in += lineWidth;
	}
}


//! converts R8G8B8 24 bit data to A1R5G5B5 data (used e.g for JPG to A1R5G5B5)
void rpColorConverter::convert24BitTo16BitColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch)
{
	const c8 *p = in;
	const i32 lineWidth = 3 * width + pitch;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (width-x-1)*3;
			*out = RGB16(*p, p[1], p[2]);
			++out;
		}

		in += lineWidth;
	}
}


//! converts R8G8B8 24 bit data to A1R5G5B5 data (used e.g for JPG to A1R5G5B5)
//! accepts colors in different order.
void rpColorConverter::convert24BitTo16BitFlipColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch)
{
	const c8 *p = in;
	const i32 lineWidth = 3 * width + pitch;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (x)*3;
			*out = RGB16(*p, p[1], p[2]);
			++out;
		}

		in += lineWidth;
	}
}


//! converts X8R8G8B8 32 bit data to A1R5G5B5 data, and flips and 
//! mirrors the image during the process.
void rpColorConverter::convert32BitTo16BitColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch)
{
	const c8 *p = in;
	const i32 lineWidth = 4 * width + pitch;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (width-x-1)*4;
			*out = RGB16(p[2], p[1], p[0]);
			++out;
		}

		in += lineWidth;
	}
}


//! converts X8R8G8B8 32 bit data to A1R5G5B5 data, and flips and 
//! mirrors the image during the process.
void rpColorConverter::convert32BitTo16BitFlipMirrorColorShuffle(const c8* in, i16* out, i32 width, i32 height, i32 pitch)
{
	const c8 *p = in;
	const i32 lineWidth = 4 * width + pitch;
	out += (width+pitch) * height;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (width-x-1)*4;
			--out;
			*out = RGB16(p[2], p[1], p[0]);
		}

		in += lineWidth;
	}
}


//! Resizes the surface to a new size and converts it at the same time
//! to an A8R8G8B8 format, returning the pointer to the new buffer.
//! The returned pointer has to be deleted.
void rpColorConverter::convert16bitToA8R8G8B8andResize(const i16* in, i32* out, i32 newWidth, i32 newHeight, i32 currentWidth, i32 currentHeight)
{
	if (!newWidth || !newHeight)
		return;

	// note: this is very very slow. (i didn't want to write a fast version.
	// but hopefully, nobody wants to convert surfaces every frame.	

	i32 nDataSize = newWidth * newHeight;
	i32 nDataSizeInBytes = nDataSize * sizeof(i32);
	//i32* nData = new i32[nDataSize];

	f32 sourceXStep = (f32)currentWidth / (f32)newWidth;
	f32 sourceYStep = (f32)currentHeight / (f32)newHeight;
	f32 sy;
	i32 t;

	for (i32 x=0; x<newWidth; ++x)
	{
		sy = 0.0f;

		for (i32 y=0; y<newHeight; ++y)
		{
			t = in[(i32)(((i32)sy)*currentWidth + x*sourceXStep)];
			t = (((t >> 15)&0x1)<<31) |	(((t >> 10)&0x1F)<<19) |
				(((t >> 5)&0x1F)<<11) |	(t&0x1F)<<3;
			out[(i32)(y*newWidth + x)] = t;

			sy+=sourceYStep;
		}
	}
}


//! copies X8R8G8B8 32 bit data, and flips and 
//! mirrors the image during the process.
void rpColorConverter::convert32BitTo32BitFlipMirror(const i32* in, i32* out, i32 width, i32 height, i32 pitch)
{
	const i32 *p = in;
	const i32 lineWidth = width;
	out += height * width;

	for (i32 y=0; y<height; ++y)
	{
		for (i32 x=0; x<width; ++x)
		{
			p = in + (width-x-1);
			--out;
			*out = *p;
		}

		in += lineWidth;
	}
}
