/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#include "include\rpimage.h"
#include <memory.h>


//! constructor
rpImage::rpImage(E_COLOR_FORMAT format, const dimension2d<i32>& size)
: Format(format), Size(size), Data(0)
{
	initData();	
}


//! constructor
rpImage::rpImage(E_COLOR_FORMAT format, const dimension2d<i32>& size, void* data)
: Format(format), Size(size), Data(0)
{
	Data = (void*)0xbadf00d;

	initData();	
	Data = data;
}




//! constructor 
rpImage::rpImage(E_COLOR_FORMAT format, rpImage* imageToCopy)
: Format(format), Data(0)
{
	if (!imageToCopy)
		return;

	Size = imageToCopy->getDimension();
	initData();

	// now copy data from other image

	if (Format == imageToCopy->getColorFormat())
	{
		memcpy(Data, imageToCopy->lock(), imageToCopy->getImageDataSizeInBytes()); 
		imageToCopy->unlock();
	}
	else
	if (Format == E_CF_A1R5G5B5)
		for (i32 x=0; x<(i32)Size.Width; ++x)
			for (i32 y=0; y<(i32)Size.Height; ++y)
				((i16*)Data)[y*Size.Width + x] = imageToCopy->getPixel(x,y).toA1R5G5B5();
	else
	{}	//unsupported format to copy from
}



//! constructor
rpImage::rpImage(rpImage* imageToCopy,	const position2d<i32>& pos,
			   const dimension2d<i32>& size)
 : Data(0), Size(0,0)
{
	if (!imageToCopy)
		return;

	Format = imageToCopy->getColorFormat();
	Size = size;

	initData();

	// copy image part
	c8* data = (c8*)Data;
	c8* source = (c8*)imageToCopy->lock();
	i32 bytes = imageToCopy->getBytesPerPixel();
	i32 srcpitch = imageToCopy->getDimension().Width * bytes;
	i32 tgtpitch = Size.Width*bytes;

	for (i32 x=0; x<size.Width; ++x)
		for (i32 y=0; y<size.Height; ++y)
		{
			for (i32 b=0; b<bytes; ++b)
				data[y*tgtpitch + (x*bytes) + b] =
					source[(y+pos.Y)*srcpitch + ((pos.X+x)*bytes) + b];
		}

	imageToCopy->unlock();
}




//! assumes format and size has been set and creates the rest
void rpImage::initData()
{
	BitsPerPixel = getBitsPerPixelFromFormat();
	BytesPerPixel = BitsPerPixel / 8;

	if (!Data)
		Data = new s8[Size.Height * Size.Width * BytesPerPixel];


}


//! destructor
rpImage::~rpImage()
{
	delete [] (s8*)Data;
}



//! Lock function.
void* rpImage::lock()
{
	return Data;
}



//! Unlock function.
void rpImage::unlock()
{
}



//! Returns width and height of image data.
const dimension2d<i32>& rpImage::getDimension()
{
	return Size;
}



//! Returns bits per pixel. 
i32 rpImage::getBitsPerPixel()
{
	return BitsPerPixel;
}


//! Returns bytes per pixel
i32 rpImage::getBytesPerPixel()
{
	return BytesPerPixel;
}



//! Returns image data size in bytes
i32 rpImage::getImageDataSizeInBytes()
{
	return BytesPerPixel * Size.Width * Size.Height;
}



//! Returns image data size in pixels
i32 rpImage::getImageDataSizeInPixels()
{
	return Size.Width * Size.Height;
}



//! returns mask for red value of a pixel
i32 rpImage::getRedMask()
{
	return RedMask;
}



//! returns mask for green value of a pixel
i32 rpImage::getGreenMask()
{
	return GreenMask;
}



//! returns mask for blue value of a pixel
i32 rpImage::getBlueMask()
{
	return BlueMask;
}



//! returns mask for alpha value of a pixel
i32 rpImage::getAlphaMask()
{
	return AlphaMask;
}


i32 rpImage::getBitsPerPixelFromFormat()
{
	switch(Format)
	{
	case E_CF_A1R5G5B5:
		AlphaMask = 0x1<<15;
		RedMask = 0x1F<<10;
		GreenMask = 0x1F<<5;
		BlueMask = 0x1F;
		return 16;
	case E_CF_R5G6B5:
		AlphaMask = 0x0;
		RedMask = 0x1F<<11;
		GreenMask = 0x3F<<5;
		BlueMask = 0x1F;
		return 16;
	case E_CF_R8G8B8:
		AlphaMask = 0x0;
		RedMask = 0xFF<<16;
		GreenMask = 0xFF<<8;
		BlueMask = 0xFF;
		return 24;
	case E_CF_A8R8G8B8:
		AlphaMask = 0xFF<<24;
		RedMask = 0xFF<<16;
		GreenMask = 0xFF<<8;
		BlueMask = 0xFF;
		return 32;
	}

	// unknown color format
	return 0;
}


//! returns a pixel
rpColor rpImage::getPixel(i32 x, i32 y)
{
	if (x < 0 || y < 0 || x >= Size.Width || y >= Size.Height)
		return rpColor(0);

	switch(Format)
	{
	case E_CF_A1R5G5B5:
		return A1R5G5B5toA8R8G8B8(((i16*)Data)[y*Size.Width + x]);
	case E_CF_R5G6B5:
		return R5G6B5toA8R8G8B8(((i16*)Data)[y*Size.Width + x]);
	case E_CF_A8R8G8B8:
		return ((i32*)Data)[y*Size.Width + x];
	case E_CF_R8G8B8:
		{
			uc8* p = &((uc8*)Data)[(y*3)*Size.Width + (x*3)];
			return rpColor(255,p[0],p[1],p[2]);
		}
	}

	return rpColor(0);
}


//! returns the color format
E_COLOR_FORMAT rpImage::getColorFormat()
{
	return Format;
}


//! clips a x coordinate into the screen
inline void rpImage::clipX(i32 &x)
{
	if (x<0)
		x = 0;
	else
		if (x > Size.Width)
			x = Size.Width;
}



//! clips a y coordinate into the screen
inline void rpImage::clipY(i32 &y)
{
	if (y<0)
		y = 0;
	else
		if (y > Size.Height)
			y = Size.Height;
}



//! exchanges two ints
inline void rpImage::exchange(i32& a, i32& b)
{
	i32 tmp;
	tmp = a;
	a = b;
	b = tmp;
}


//! draws a rectangle
void rpImage::drawRectangle(const rect<i32>& rect, rpColor color)
{
	drawRectangle(	rect.UpperLeftCorner.X, rect.UpperLeftCorner.Y, 
					rect.LowerRightCorner.X, rect.LowerRightCorner.Y, color);
}


//! draws a rectangle
void rpImage::drawRectangle(i32 x, i32 y, i32 x2, i32 y2, rpColor color)
{
	if (Format != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5
		return;
	}

	// clip

	clipX(x); clipX(x2);
	clipY(y); clipY(y2);

	// switch x and x2 if neccecary

	if (x > x2)
		exchange(x,x2);

	if (y > y2)
		exchange(y,y2);

	// draw

	i32 l=y*Size.Width;
	i32 ix;

	if (color.getAlpha()==255)
	{
		// quickly draw without alpha.

		i16 c = color.toA1R5G5B5();

		for (i32 iy=y; iy<y2; ++iy)
		{
			for (ix=x; ix<x2; ++ix)
				((i16*)Data)[l + ix] = c;

			l += Size.Width;
		}
	}
	else
	{
		// draw with alpha

		i32 ia = color.getAlpha();
		i32 a = 255-ia;

		i32 r = getRed(color.toA1R5G5B5()) * ia;
		i32 g = getGreen(color.toA1R5G5B5()) * ia;
		i32 b = getBlue(color.toA1R5G5B5()) * ia;

		i16 *src;

		for (i32 iy=y; iy<y2; ++iy)
		{
			for (ix=x; ix<x2; ++ix)
			{
				src = &((i16*)Data)[l + ix];
				*src =	RGB16(	
					(getRed(*src)*a + r)>>5,
					(getGreen(*src)*a + g)>>5,
					(getBlue(*src)*a + b)>>5);
			}

			l += Size.Width;
		}
	}
}



//! clips coordinates into the rect. returns false if the rects is completely outside
//! and does not need to be drawn.
inline bool rpImage::clipRect(const i32 targetWidth, const i32 targetHeight, i32& targetX, i32& targetY, i32& xInSource, i32& yInSource, i32& sourceWidth, i32& soureHeight)
{
	if (targetX<0)
	{
		sourceWidth += targetX;
		if (sourceWidth <= 0)
			return false;

		xInSource -= targetX;
		targetX = 0;
	}
	
	if (targetX+sourceWidth>targetWidth)
	{
		sourceWidth -= (targetX + sourceWidth) - targetWidth;
		if (sourceWidth <= 0)
			return false;
	}

	if (targetY<0)
	{
		soureHeight += targetY;
		if (soureHeight <= 0)
			return false;

		yInSource -= targetY;
		targetY = 0;
	}
	
	if (targetY+soureHeight>targetHeight)
	{
		soureHeight -= (targetY + soureHeight) - targetHeight;
		if (soureHeight <= 0)
			return false;
	}

	return true;
}



//! copies this surface into another
void rpImage::copyTo(rpImage* target, const position2d<i32>& pos)
{
	copyTo(target, pos.X, pos.Y);
}



//! copies this surface into another
void rpImage::copyTo(rpImage* target, i32 x, i32 y)
{
	if (Format != E_CF_A1R5G5B5 ||
		target->getColorFormat() != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5
		return;
	}

	i16* data = (i16*)target->lock();
	dimension2d<i32> size = target->getDimension();

	// clip

	i32 ownWidth = Size.Width;
	i32 ownHeight = Size.Height;
	i32 ownX = 0;
	i32 ownY = 0;

	if (!clipRect(size.Width, size.Height, x, y, ownX, ownY, ownWidth, ownHeight))
		return;

	// copy

	i32 ltarget=y*size.Width + x;
	i32 lown=ownY*Size.Width + ownX;
	size_t dataRowSizeInBytes = ownWidth*sizeof(i16);

	for (i32 iy=0; iy<ownHeight; ++iy)
	{
		memcpy(&data[ltarget], &((i16*)Data)[lown], dataRowSizeInBytes); 

		lown += Size.Width;
		ltarget += size.Width;
	}

	target->unlock();
}



//! copies this surface into another
void rpImage::copyTo(rpImage* target, const position2d<i32>& pos, const rect<i32>& sourceRect, const rect<i32>* clipRect)
{
	if (Format != E_CF_A1R5G5B5 ||
		target->getColorFormat() != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5 
		return;
	}

	if (sourceRect.UpperLeftCorner.X >= sourceRect.LowerRightCorner.X ||
		sourceRect.UpperLeftCorner.Y >= sourceRect.LowerRightCorner.Y)
		return;

	position2d<i32> targetPos = pos;
	position2d<i32> sourcePos = sourceRect.UpperLeftCorner;
	dimension2d<i32> sourceSize(sourceRect.getWidth(), sourceRect.getHeight());
	const dimension2d<i32> targetSurfaceSize = target->getDimension();

	if (clipRect)
	{
		if (targetPos.X < clipRect->UpperLeftCorner.X)
		{
			sourceSize.Width += targetPos.X - clipRect->UpperLeftCorner.X;
			if (sourceSize.Width <= 0)
				return;

			sourcePos.X -= targetPos.X - clipRect->UpperLeftCorner.X;
			targetPos.X = clipRect->UpperLeftCorner.X;
		}
		
		if (targetPos.X + sourceSize.Width > clipRect->LowerRightCorner.X)
		{
			sourceSize.Width -= (targetPos.X + sourceSize.Width) - clipRect->LowerRightCorner.X;
			if (sourceSize.Width <= 0)
				return;
		}

		if (targetPos.Y < clipRect->UpperLeftCorner.Y)
		{
			sourceSize.Height += targetPos.Y - clipRect->UpperLeftCorner.Y;
			if (sourceSize.Height <= 0)
				return;

			sourcePos.Y -= targetPos.Y - clipRect->UpperLeftCorner.Y;
			targetPos.Y = clipRect->UpperLeftCorner.Y;
		}
		
		if (targetPos.Y + sourceSize.Height > clipRect->LowerRightCorner.Y)
		{
			sourceSize.Height -= (targetPos.Y + sourceSize.Height) - clipRect->LowerRightCorner.Y;
			if (sourceSize.Height <= 0)
				return;
		}
	}

	// clip these coordinates

	if (targetPos.X<0)
	{
		sourceSize.Width += targetPos.X;
		if (sourceSize.Width <= 0)
			return;

		sourcePos.X -= targetPos.X;
		targetPos.X = 0;
	}
	
	if (targetPos.X + sourceSize.Width > targetSurfaceSize.Width)
	{
		sourceSize.Width -= (targetPos.X + sourceSize.Width) - targetSurfaceSize.Width;
		if (sourceSize.Width <= 0)
			return;
	}

	if (targetPos.Y<0)
	{
		sourceSize.Height += targetPos.Y;
		if (sourceSize.Height <= 0)
			return;

		sourcePos.Y -= targetPos.Y;
		targetPos.Y = 0;
	}
	
	if (targetPos.Y + sourceSize.Height > targetSurfaceSize.Height)
	{
		sourceSize.Height -= (targetPos.Y + sourceSize.Height) - targetSurfaceSize.Height;
		if (sourceSize.Height <= 0)
			return;
	}

	// draw everything

	i16* targetData = (i16*)target->lock();
	i32 ltarget = targetPos.Y * targetSurfaceSize.Width + targetPos.X;
	i32 lsource = sourcePos.Y * Size.Width + sourcePos.X;
	size_t dataRowSizeInBytes = sourceSize.Width * sizeof(i16);

	for (i32 iy=0; iy<sourceSize.Height; ++iy)
	{
		memcpy(&targetData[ltarget], &((i16*)Data)[lsource], dataRowSizeInBytes); 
		lsource += Size.Width;
		ltarget += targetSurfaceSize.Width;
	}
}



//! copies this surface into another, using the alpha mask
void rpImage::copyToWithAlpha(rpImage* target, const position2d<i32>& pos, const rect<i32>& sourceRect)
{
	if (Format != E_CF_A1R5G5B5 ||
		target->getColorFormat() != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5 
		return;
	}


	if (sourceRect.UpperLeftCorner.X >= sourceRect.LowerRightCorner.X ||
		sourceRect.UpperLeftCorner.Y >= sourceRect.LowerRightCorner.Y)
		return;

	position2d<i32> targetPos = pos;
	position2d<i32> sourcePos = sourceRect.UpperLeftCorner;
	dimension2d<i32> sourceSize(sourceRect.getWidth(), sourceRect.getHeight());
	const dimension2d<i32> targetSurfaceSize = target->getDimension();

	// clip these coordinates

	if (targetPos.X<0)
	{
		sourceSize.Width += targetPos.X;
		if (sourceSize.Width <= 0)
			return;

		sourcePos.X -= targetPos.X;
		targetPos.X = 0;
	}
	
	if (targetPos.X + sourceSize.Width > targetSurfaceSize.Width)
	{
		sourceSize.Width -= (targetPos.X + sourceSize.Width) - targetSurfaceSize.Width;
		if (sourceSize.Width <= 0)
			return;
	}

	if (targetPos.Y<0)
	{
		sourceSize.Height += targetPos.Y;
		if (sourceSize.Height <= 0)
			return;

		sourcePos.Y -= targetPos.Y;
		targetPos.Y = 0;
	}
	
	if (targetPos.Y + sourceSize.Height > targetSurfaceSize.Height)
	{
		sourceSize.Height -= (targetPos.Y + sourceSize.Height) - targetSurfaceSize.Height;
		if (sourceSize.Height <= 0)
			return;
	}

	// draw everything

	i16* targetData = (i16*)target->lock();
	i32 ltarget = targetPos.Y * targetSurfaceSize.Width + targetPos.X;
	i32 lsource = sourcePos.Y * Size.Width + sourcePos.X;
	size_t dataRowSizeInBytes = sourceSize.Width * sizeof(i16);

	i16 alphaMask = getAlphaMask();
	i16* p;
	i16* end;
	i16* t;

	for (i32 iy=0; iy<sourceSize.Height; ++iy)
	{
		p = &((i16*)Data)[lsource];
		end = p + (dataRowSizeInBytes>>1);
		t = &targetData[ltarget];

		while(p != end)
		{
			if (*p & alphaMask)
				*t = *p;

			++t;
			++p;
		}
		lsource += Size.Width;
		ltarget += targetSurfaceSize.Width;
	}

}




//! copies this surface into another, using the alpha mask, an cliprect and a color to add with
void rpImage::copyToWithAlpha(rpImage* target, const position2d<i32>& pos, const rect<i32>& sourceRect, i16 color, const rect<i32>* clipRect)
{
	if (Format != E_CF_A1R5G5B5 ||
		target->getColorFormat() != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5
		return;
	}

	if (sourceRect.UpperLeftCorner.X >= sourceRect.LowerRightCorner.X ||
		sourceRect.UpperLeftCorner.Y >= sourceRect.LowerRightCorner.Y)
		return;

	position2d<i32> targetPos = pos;
	position2d<i32> sourcePos = sourceRect.UpperLeftCorner;
	dimension2d<i32> sourceSize(sourceRect.getWidth(), sourceRect.getHeight());
	const dimension2d<i32> targetSurfaceSize = target->getDimension();

	// clip to cliprect if there is one
	if (clipRect)
	{
		if (targetPos.X < clipRect->UpperLeftCorner.X)
		{
			sourceSize.Width += targetPos.X - clipRect->UpperLeftCorner.X;
			if (sourceSize.Width <= 0)
				return;

			sourcePos.X -= targetPos.X - clipRect->UpperLeftCorner.X;
			targetPos.X = clipRect->UpperLeftCorner.X;
		}
		
		if (targetPos.X + sourceSize.Width > clipRect->LowerRightCorner.X)
		{
			sourceSize.Width -= (targetPos.X + sourceSize.Width) - clipRect->LowerRightCorner.X;
			if (sourceSize.Width <= 0)
				return;
		}

		if (targetPos.Y < clipRect->UpperLeftCorner.Y)
		{
			sourceSize.Height += targetPos.Y - clipRect->UpperLeftCorner.Y;
			if (sourceSize.Height <= 0)
				return;

			sourcePos.Y -= targetPos.Y - clipRect->UpperLeftCorner.Y;
			targetPos.Y = clipRect->UpperLeftCorner.Y;
		}
		
		if (targetPos.Y + sourceSize.Height > clipRect->LowerRightCorner.Y)
		{
			sourceSize.Height -= (targetPos.Y + sourceSize.Height) - clipRect->LowerRightCorner.Y;
			if (sourceSize.Height <= 0)
				return;
		}
	}

	// clip these coordinates to screen

	if (targetPos.X<0)
	{
		sourceSize.Width += targetPos.X;
		if (sourceSize.Width <= 0)
			return;

		sourcePos.X -= targetPos.X;
		targetPos.X = 0;
	}
	
	if (targetPos.X + sourceSize.Width > targetSurfaceSize.Width)
	{
		sourceSize.Width -= (targetPos.X + sourceSize.Width) - targetSurfaceSize.Width;
		if (sourceSize.Width <= 0)
			return;
	}

	if (targetPos.Y<0)
	{
		sourceSize.Height += targetPos.Y;
		if (sourceSize.Height <= 0)
			return;

		sourcePos.Y -= targetPos.Y;
		targetPos.Y = 0;
	}
	
	if (targetPos.Y + sourceSize.Height > targetSurfaceSize.Height)
	{
		sourceSize.Height -= (targetPos.Y + sourceSize.Height) - targetSurfaceSize.Height;
		if (sourceSize.Height <= 0)
			return;
	}

	// draw everything

	i16* targetData = (i16*)target->lock();
	i32 ltarget = targetPos.Y * targetSurfaceSize.Width + targetPos.X;
	i32 lsource = sourcePos.Y * Size.Width + sourcePos.X;
	size_t dataRowSizeInBytes = sourceSize.Width * sizeof(i16);

	i16 alphaMask = getAlphaMask();
	i16* p;
	i16* end;
	i16* t;

	i32 r = getRed(color);
	i32 g = getGreen(color);
	i32 b = getBlue(color);

	for (i32 iy=0; iy<sourceSize.Height; ++iy)
	{
		p = &((i16*)Data)[lsource];
		end = p + (dataRowSizeInBytes>>1);
		t = &targetData[ltarget];

		while(p != end)
		{
			if (*p & alphaMask)
				*t = RGB16(getRed(*p) * (r) >>2, getGreen(*p) * (g) >>2, getBlue(*p) * (b) >>2);

			++t;
			++p;
		}
		lsource += Size.Width;
		ltarget += targetSurfaceSize.Width;
	}
}



//! draws a line from to
void rpImage::drawLine(const position2d<i32>& from, const position2d<i32>& to, i16 color)
{
	if (Format != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5
		return;
	}

	i32 x = from.X;
    i32 y = from.Y;

	i32 deltax = to.X - from.X;
	i32 deltay = to.Y - from.Y;
 
	i32 stepx = deltax < 0 ? -1 : 1;
	i32 stepy = deltay < 0 ? -1 : 1;
 
	deltax *= stepx; 
	deltay *= stepy; 
 
	if (deltax > deltay)
	{
		 i32 d = deltax >> 1;
 
		 while (x != to.X)
		 {
			if (x>=0 && y>=0 && x<Size.Width && y<Size.Height)
				setPixelFast(x, y, color);

			d -= deltay;

			if (d < 0)
			{
				y += stepy;
				d += deltax;
			}

			x += stepx;       
		 }
	}
	else
	{
		 i32 d = deltay >> 1;
 
		 while (y != to.Y)
		 {
			if (x>=0 && y>=0 && x<Size.Width && y<Size.Height)
				setPixelFast(x, y, color);

			d -= deltax;
			if (d < 0)
			{
				x += stepx;
				d += deltay;
			}

			y += stepy;
		 }
	}
}


//! sets a pixel very fast and inline
inline void rpImage::setPixelFast(i32 x, i32 y, i16 color)
{
	((i16*)Data)[y*Size.Width + x] = color;
}


//! resizes the surface to a new size
void rpImage::resizeTo(const dimension2d<i32>& size)
{
	if (Format != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5
		return;
	}

	// note: this is very very slow. (i didn't want to write a fast version.
	// but hopefully, nobody wants to scale surfaces every frame.

	if (!size.Width || !size.Height)
		return;

	i32 nDataSize = size.Width * size.Height;
	i32 nDataSizeInBytes = nDataSize * sizeof(i16);
	i16* nData = new i16[nDataSize];

	f32 sourceXStep = (f32)Size.Width / (f32)size.Width;
	f32 sourceYStep = (f32)Size.Height / (f32)size.Height;
	f32 sy;

    for (i32 x=0; x<size.Width; ++x)
	{
		sy = 0.0f;

		for (i32 y=0; y<size.Height; ++y)
		{
			nData[(i32)(y*size.Width + x)] = ((i16*)Data)[(i32)(((i32)sy)*Size.Width + x*sourceXStep)];
			sy+=sourceYStep;
		}
	}

	delete [] (s8*)Data;
    //DataSize = nDataSize;
	//DataSizeInBytes = nDataSizeInBytes;
	Size = size;
	Data = nData;
}



//! copies this surface into another, scaling it to fit it.
void rpImage::copyToScaling(rpImage* target)
{
	if (Format != E_CF_A1R5G5B5 ||
		target->getColorFormat() != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5
		return;
	}

	// note: this is very very slow. (i didn't want to write a fast version.
	// but hopefully, nobody wants to scale surfaces every frame.

	dimension2d<i32> size = target->getDimension();

	i32 nDataSize = size.Width * size.Height;
	i32 nDataSizeInBytes = nDataSize * sizeof(i16);

	if (!size.Width || !size.Height)
		return;

	i16* nData = (i16*)target->lock();

	f32 sourceXStep = (f32)Size.Width / (f32)size.Width;
	f32 sourceYStep = (f32)Size.Height / (f32)size.Height;
	f32 sy;

    for (i32 x=0; x<size.Width; ++x)
	{
		sy = 0.0f;

		for (i32 y=0; y<size.Height; ++y)
		{
			nData[(i32)(y*size.Width + x)] = ((i16*)Data)[(i32)(((i32)sy)*Size.Width + x*sourceXStep)];
			sy+=sourceYStep;
		}
	}

	target->unlock();
}


//! fills the surface with black or white
void rpImage::fill(i16 color)
{
	if (Format != E_CF_A1R5G5B5)
	{
		// rpImage is not A1R5G5B5
		return;
	}


	i32* p = (i32*)Data;
	i32* bufferEnd = p + ((Size.Width * Size.Height)>>1);
	i32 c = ((color & 0x0000ffff)<<16) | (color & 0x0000ffff);
	while(p != bufferEnd)
	{
		*p = c;
		++p;
	}
}