/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


// Copyright (C) 2002-2005 Nikolaus Gebhardt


#include "rpLoaderBMP.h"
#include <string.h>
#include "rpColor.h"
//#include "rpColorConverter.h"
#include "rpImage.h"


//! constructor
rpLoaderBMP::rpLoaderBMP()
: BmpData(0), PaletteData(0)
{
	#ifdef _DEBUG
	setDebugName("rpLoaderBMP");
	#endif
}



//! destructor
rpLoaderBMP::~rpLoaderBMP()
{
	if (PaletteData)
		delete [] PaletteData;

	if (BmpData)
		delete [] BmpData;
}



//! returns true if the file maybe is able to be loaded by this class
//! based on the file extension (e.g. ".tga")
bool rpLoaderBMP::isALoadableFileExtension(const c8* fileName)
{
	return strstr(fileName, ".bmp") != 0;
}



//! returns true if the file maybe is able to be loaded by this class
bool rpLoaderBMP::isALoadableFileFormat(FILE* file)
{
	ui16 headerID;
	fread(&headerID, sizeof(ui16),1,file);
	return headerID == 0x4d42;
}


void rpLoaderBMP::decompress8BitRLE(c8*& BmpData, i32 size, i32 width, i32 height, i32 pitch)
{
	c8* p = BmpData;
	c8* newBmp = new c8[(width+pitch)*height];
	c8* d = newBmp;
	c8* destEnd = newBmp + (width+pitch)*height;
	i32 line = 0;

	while (BmpData - p < size && d < destEnd)
	{
		if (*p == 0)
		{
			++p;

			switch(*p)
			{
			case 0: // end of line
				++p; 
				++line;
				d = newBmp + (line*(width+pitch));
				break; 
			case 1: // end of bmp
				delete [] BmpData;
				BmpData = newBmp;
				return; 
			case 2:
				++p; d +=(uc8)*p;  // delta
				++p; d += ((uc8)*p)*(width+pitch); 
				++p;
				break;
			default:
				{
					// absolute mode
					i32 count = (uc8)*p; ++p;
					i32 readAdditional = ((2-(count%2))%2);
					i32 i;

					for (i=0; i<count; ++i)
					{
						*d = *p;
						++p;
						++d;
					}

					for (i=0; i<readAdditional; ++i)
						++p;
				}
			}
		}
		else
		{
			i32 count = (uc8)*p; ++p;			
			i32 color = (uc8)*p; ++p;
			for (i32 i=0; i<count; ++i)
			{
				*d = color;
				++d;
			}
		}
	}

	delete [] BmpData;
	BmpData = newBmp;
}


void rpLoaderBMP::decompress4BitRLE(c8*& BmpData, i32 size, i32 width, i32 height, i32 pitch)
{
	i32 lineWidth = (width+1)/2+pitch;
	c8* p = BmpData;
	c8* newBmp = new c8[lineWidth*height];
	c8* d = newBmp;
	c8* destEnd = newBmp + lineWidth*height;
	i32 line = 0;
	i32 shift = 4;

	while (BmpData - p < size && d < destEnd)
	{
		if (*p == 0)
		{
			++p;

			switch(*p)
			{
			case 0: // end of line
				++p; 
				++line;
				d = newBmp + (line*lineWidth);
				shift = 4;
				break; 
			case 1: // end of bmp
				delete [] BmpData;
				BmpData = newBmp;
				return; 
			case 2:
				{
					++p; 
					i32 x = (uc8)*p; ++p;
					i32 y = (uc8)*p; ++p;
					d += x/2 + y*lineWidth;
					shift = x%2==0 ? 4 : 0;
				}
				break;
			default:
				{
					// absolute mode
					i32 count = (uc8)*p; ++p;
					i32 readAdditional = ((2-((count)%2))%2);
					i32 readShift = 4;
					i32 i;

					for (i=0; i<count; ++i)
					{
						i32 color = (((uc8)*p) >> readShift) & 0x0f;
						readShift -= 4;
						if (readShift < 0)
						{
							++*p;
							readShift = 4;
						}

						c8 mask = 0x0f << shift;
						*d = (*d & (~mask)) | ((color << shift) & mask);

						shift -= 4;
						if (shift < 0)
						{
							shift = 4;
							++d;
						}

					}

					for (i=0; i<readAdditional; ++i)
						++p;
				}
			}
		}
		else
		{
			i32 count = (uc8)*p; ++p;			
			i32 color1 = (uc8)*p; color1 = color1 & 0x0f;
			i32 color2 = (uc8)*p; color2 = (color2 >> 4) & 0x0f;
			++p;

			for (i32 i=0; i<count; ++i)
			{
				c8 mask = 0x0f << shift;
				c8 toSet = (shift==0 ? color1 : color2) << shift;
				*d = (*d & (~mask)) | (toSet & mask);

				shift -= 4;
				if (shift < 0)
				{
					shift = 4;
					++d;
				}
			}
		}
	}

	delete [] BmpData;
	BmpData = newBmp;
}



//! creates a surface from the file
rpImage* rpLoaderBMP::loadImage(const c8* fileN)
{
	if (!fileN) return NULL;

	FILE *file = fopen(fileN, "rb");
	if (!file) return NULL;
	
	SBMPHeader header;

	fseek(file, 0 , SEEK_END);
	i32 fsize = ftell(file);
	fseek(file, 0, SEEK_SET);

	fread(&header, sizeof(header),1,file);

	i32 pitch = 0;

	//! return if the header is false

	if (header.Id != 0x4d42 && header.Id != 0x424d)
		return 0;

	if (header.Compression != 0)
	{
		// compressed not supported yet
		return 0;
	}

	// adjust bitmap data size to dword boundary
	header.BitmapDataSize += (4-(header.BitmapDataSize%4))%4;

	// read palette

	i32 pos = ftell(file);
	i32 paletteSize = (header.BitmapDataOffset - pos) / 4;

	if (paletteSize)
	{
		PaletteData = new i32[paletteSize];
		fread(PaletteData, paletteSize * sizeof(i32), 1, file);
	}

	// read image data

	if (!header.BitmapDataSize)
	{
		// okay, wir drfen raten wie gross die bitmap ist.
		// manche Ar*******er speichern das nicht ab.
		header.BitmapDataSize = fsize - header.BitmapDataOffset;
	}


	fseek(file, header.BitmapDataOffset, SEEK_SET);

	f32 t = (header.Width) * (header.BPP / 8.0f);
	i32 widthInBytes = (i32)t;
	t -= widthInBytes;
	if (t!=0.0f)
		++widthInBytes;

	i32 lineData = widthInBytes + ((4-(widthInBytes%4)))%4;
	pitch = lineData - widthInBytes;
	
	BmpData = new c8[header.BitmapDataSize];
	fread(BmpData, header.BitmapDataSize, 1, file);


	// decompress data if needed
	switch(header.Compression)
	{
	case 1: // 8 bit rle
		decompress8BitRLE(BmpData, header.BitmapDataSize, header.Width, header.Height, pitch);
		break;
	case 2: // 4 bit rle
		decompress4BitRLE(BmpData, header.BitmapDataSize, header.Width, header.Height, pitch);
		break;
	}

	// create surface

	rpImage* image = 0;

	switch(header.BPP)
	{
	case 1:
		image = new rpImage(E_CF_A1R5G5B5, dimension2d<i32>(header.Width, header.Height));
		rpColorConverter::convert1BitTo16BitFlipMirror(BmpData, (i16*)image->lock(), header.Width, header.Height, pitch);
		image->unlock();
		break;
	case 4:
		image = new rpImage(E_CF_A1R5G5B5, dimension2d<i32>(header.Width, header.Height));
		rpColorConverter::convert4BitTo16BitFlipMirror(BmpData, (i16*)image->lock(), header.Width, header.Height, pitch, PaletteData);
		image->unlock();
		break;
	case 8:
		image = new rpImage(E_CF_A1R5G5B5, dimension2d<i32>(header.Width, header.Height));
		rpColorConverter::convert8BitTo16BitFlipMirror(BmpData, (i16*)image->lock(), header.Width, header.Height, pitch, PaletteData);
		image->unlock();
		break;
	case 16:
		break;
	case 24:
		image = new rpImage(E_CF_R8G8B8, dimension2d<i32>(header.Width, header.Height));
		rpColorConverter::convert24BitTo24BitFlipMirrorColorShuffle(BmpData, (c8*)image->lock(), header.Width, header.Height, pitch);
		image->unlock();
		break;
	case 32: // thx to Reinhard Ostermeier
		image = new rpImage(E_CF_A8R8G8B8, dimension2d<i32>(header.Width, header.Height));
		rpColorConverter::convert32BitTo32BitFlipMirror((i32*)BmpData, (i32*)image->lock(), header.Width, header.Height, pitch);
		image->unlock();
		break;
	};

	// clean up

	delete [] PaletteData;
	PaletteData = 0;

	delete [] BmpData;
	BmpData = 0;

	return image;
}


//! creates a loader which is able to load windows bitmaps
rpImageLoader* createImageLoaderBmp()
{
	return new rpLoaderBMP;
}

