/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#include "include\rpphysicsCollision.h"
#include "rpGeometryBuffer.h"
#include "rpVector.h"

rpPhysicsCollision::rpPhysicsCollision(rpPhysicsWorld *pW,
									   E_PCOLLISION_TYPE type,
									   f32 size_x,
									   f32 size_y,
									   f32 size_z,
									   rpMatrix *offset):pWorld(pW)
{
	collision_type = E_PCT_NONE;
	if (!pWorld)
		return;
	
	switch (type) {
		case E_PCT_BOX:
			{
	            pCollision = NewtonCreateBox(pWorld->nWorld, size_x, size_y, size_z, (f32*)offset);
				collision_type = type;
				break;
			}
		case E_PCT_SPHERE:
			{
				pCollision = NewtonCreateSphere(pWorld->nWorld, size_x, size_y, size_z, (f32*)offset);
				collision_type = type;
				break;
			}
		default:
			break;	// invalid type
	}

}
rpPhysicsCollision::rpPhysicsCollision(rpPhysicsWorld *pW,
									   E_PCOLLISION_TYPE type,
									   f32 radius,
									   f32 height,
									   rpMatrix *offset):pWorld(pW), pCollision(NULL)
{
	collision_type = E_PCT_NONE;
	if (!pWorld)
		return;
	
	switch (type) {
		case E_PCT_CONE:
			{
	            pCollision = NewtonCreateCone(pWorld->nWorld, radius, height, (f32*)offset);
				collision_type = type;
				break;
			}
		case E_PCT_CAPSULE:
			{
				pCollision = NewtonCreateCapsule(pWorld->nWorld, radius, height, (f32*)offset);
				collision_type = type;
				break;
			}
		case E_PCT_CYLINDER:
			{
				pCollision = NewtonCreateCylinder(pWorld->nWorld, radius, height, (f32*)offset);
				collision_type = type;
				break;
			}
		default:
			break;	// invalid type
	}

}
void TreeCollisionCallback (
	const NewtonBody* bodyWithTreeCollision, 
	const NewtonBody* body,
	const float* vertex, 
	int vertexstrideInBytes, 
	int indexCount, 
	const int* indexArray)
{
	

}

rpPhysicsCollision::rpPhysicsCollision(rpPhysicsWorld *pW,
									   E_PCOLLISION_TYPE type,
									   rpGeometryBuffer *mesh,
									   rpMatrix *offset):pWorld(pW), pCollision(NULL)
{
	if (!pWorld)
		return;
	if (mesh) {
		collision_type = type;

		switch (type) {
		case E_PCT_STATIC_GEOMETRY:
			{
				// create the collsion tree geometry
				pCollision = NewtonCreateTreeCollision(pWorld->nWorld, TreeCollisionCallback);

				// prepare to create collsion geometry
				NewtonTreeCollisionBeginBuild(pCollision);

				switch (mesh->type) {
				case E_GB_PLAIN:
					{
						// load non-indexed mesh buffer
						i32 size = mesh->getVertexCount();
						for (i32 face = 0; face < (size / 3); face++) {
							sStandard3f face_v1 = mesh->getVertex(face * 3);
							sStandard3f face_v2 = mesh->getVertex((face * 3) + 1);
							sStandard3f face_v3 = mesh->getVertex((face * 3) + 2);
							
							// build face definition
							f32 face_def[sizeof(f32) * 9]; // 3 floats in vertex, 3 vertices
							face_def[0] = face_v1.x;face_def[1] = face_v1.y;face_def[2] = face_v1.z;
							face_def[3] = face_v2.x;face_def[4] = face_v2.y;face_def[5] = face_v2.z;
							face_def[6] = face_v3.x;face_def[7] = face_v3.y;face_def[8] = face_v3.z;

							NewtonTreeCollisionAddFace(pCollision, 3, face_def, sizeof(f32) * 3, 1);
							
						}
							
					}

				}

				// finish the geometry, (1 - optimize)
				NewtonTreeCollisionEndBuild(pCollision, 1);

				return;
			}
        case E_PCT_CONVEX_HULL:
			{
				switch (mesh->type) {
				case E_GB_PLAIN:
					{
						f32 *hull_def = new f32[mesh->getVertexCount() * 3];
						for (int i = 0; i < mesh->getVertexCount(); i++) {
							sStandard3f v = mesh->getVertex(i);
							hull_def[i * 3] = v.x;
							hull_def[(i * 3) + 1] = v.y;
							hull_def[(i * 3) + 2] = v.z;
						}
						
						// create the collsion geometry
						pCollision = NewtonCreateConvexHull(pWorld->nWorld, mesh->getVertexCount(), hull_def, sizeof(f32) * 3, (f32*)offset);
				
					}

				}
				break;
			}
		}
	}

}

rpPhysicsCollision::~rpPhysicsCollision(void)
{
	if (pWorld && pCollision)
		NewtonReleaseCollision(pWorld->nWorld, pCollision);
}


bool rpPhysicsCollision::isFine()
{
	return ((collision_type != E_PCT_NONE) && (pCollision != NULL));
}

NewtonCollision*	rpPhysicsCollision::getNewtonCollision()
{
	return pCollision;
}