/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#include "include\rpphysicsnode.h"
#include "rpSceneNode.h"

// set the tranformation of a rigid body
void _cdecl PhysicsApplyForceAndTorque (const NewtonBody* body)
{
	rpPhysicsNode *node = (rpPhysicsNode*) NewtonBodyGetUserData (body);
	if (node)
		node->applyForceAndTorque();

}

// add force and torque to rigid body
void _cdecl PhysicsSetTransform (const NewtonBody* body, const f32* matrix)
{
	rpPhysicsNode *node = (rpPhysicsNode*) NewtonBodyGetUserData (body);
	if (node)
		node->applyTransform((rpMatrix*)matrix);
}

rpPhysicsNode::rpPhysicsNode(rpPhysicsWorld *pW, rpPhysicsCollision *pC):pWorld(pW), nBody(NULL), affected_node(NULL)
{
	if (pW && pC)
		if (pC->isFine()) {
			nBody = NewtonCreateBody(pW->nWorld, pC->getNewtonCollision());

			if (nBody) {

				if (pC->collision_type < E_PCT_STATIC_GEOMETRY) {
					// save the pointer to the object with the body.
					NewtonBodySetUserData (nBody, this);
				
					// set the tranform call back function
					NewtonBodySetTransformCallback (nBody, PhysicsSetTransform);

					// set the force and torque call back funtion
					NewtonBodySetForceAndTorqueCallback (nBody, PhysicsApplyForceAndTorque);

					NewtonBodySetMassMatrix (nBody, 1.0f, 1.0f, 1.0f, 1.0f);
					
					// unfreeze object
					NewtonBodySetAutoFreeze(nBody, 0);
				}

			}
		}

		weight = 1.0;
}

rpPhysicsNode::~rpPhysicsNode(void)
{
	if (pWorld && isFine())
		NewtonDestroyBody(pWorld->nWorld, nBody);
	
	if (affected_node)
		delete affected_node;

}

void rpPhysicsNode::setSceneNode(rpSceneNode *sN)
{
	if (sN) {
		affected_node = sN;
		sN->grab();
	}
}

bool rpPhysicsNode::isFine()
{
	return (nBody != NULL);
}

// 'force' methods
void rpPhysicsNode::applyForceAndTorque()
{
	// do some simple gravity
	f32 mass, obsolete;
	getMassMatrix(mass, obsolete, obsolete, obsolete);
	
	// negate the global interia and apply weight factor
	f32 g_force = -mass * 9.8 * weight;

	rpVector3D	force(0.0,g_force, 0.0);
	addForce(force);

}
void rpPhysicsNode::applyTransform(rpMatrix *mat)
{
	if (mat && affected_node) {
		
		mat->copyTo(affected_node->transformation);
		
		// to avoid updating from vector transformation
		// since physicsNode cares about transfrmation of sceneNode, all transformation vectors are invalid
		affected_node->transformation_changed = false;
	}
}

// NEWTON methods
// interface to newton

// mass matrix
void rpPhysicsNode::setMassMatrix(f32 mass, f32 Ixx, f32 Iyy,f32 Izz)
{
	if (isFine())
		NewtonBodySetMassMatrix(nBody, mass, Ixx,Iyy, Izz);
}
void rpPhysicsNode::getMassMatrix(f32& mass, f32& Ixx, f32& Iyy,f32& Izz)
{
	if (isFine())
		NewtonBodyGetMassMatrix(nBody, &mass, &Ixx,&Iyy, &Izz);
}

// matrix
void rpPhysicsNode::setMatrix(rpMatrix *M)
{
	if (isFine()) {
		NewtonBodySetMatrix(nBody, (f32 *)M);
		applyTransform(M);
	}
}
void rpPhysicsNode::getMatrix(rpMatrix *M)
{
	if (isFine())
		NewtonBodyGetMatrix(nBody, (f32*)M);
}

// force
void rpPhysicsNode::setForce(rpVector3D &force)
{
	if (isFine())
		NewtonBodySetForce(nBody, (f32 *)&force);
}
void rpPhysicsNode::addForce(rpVector3D &force)
{
	if (isFine())
		NewtonBodyAddForce(nBody, (f32 *)&force);
}
void rpPhysicsNode::getForce(rpVector3D &force)
{
	if (isFine()) 
		NewtonBodyGetForce(nBody, (f32 *)&force);
}

// torque
void rpPhysicsNode::setTorque(rpVector3D &torque)
{
	if (isFine())
		NewtonBodySetTorque(nBody, (f32 *)&torque);
}
void rpPhysicsNode::addTorque(rpVector3D &torque)
{
	if (isFine())
		NewtonBodyAddTorque(nBody, (f32 *)&torque);
}
void rpPhysicsNode::getTorque(rpVector3D &torque)
{
	if (isFine()) 
		NewtonBodyGetTorque(nBody, (f32 *)&torque);
}

// velocity
void rpPhysicsNode::setVelocity(rpVector3D &velocity)
{
	if (isFine())
		NewtonBodySetVelocity(nBody, (f32 *)&velocity);
}
void rpPhysicsNode::getVelocity(rpVector3D &velocity)
{
	if (isFine()) 
		NewtonBodyGetVelocity(nBody, (f32 *)&velocity);
}

// omega
void rpPhysicsNode::setOmega(rpVector3D &omega)
{
	if (isFine())
		NewtonBodySetOmega(nBody, (f32 *)&omega);
}
void rpPhysicsNode::getOmega(rpVector3D &omega)
{
	if (isFine()) 
		NewtonBodyGetOmega(nBody, (f32 *)&omega);
}

// linear damping
void rpPhysicsNode::setLinearDamping(const f32 damp)
{
	if (isFine())
		NewtonBodySetLinearDamping(nBody, damp);
}
f32	rpPhysicsNode::getLinearDamping()
{
	if (isFine()) 
		return NewtonBodyGetLinearDamping(nBody);
	else return -1.0;
}

// angular damping
void rpPhysicsNode::setAngularDamping(rpVector3D &damp)
{
	if (isFine())
		NewtonBodySetAngularDamping(nBody, (f32*)&damp);
}
void rpPhysicsNode::getAngularDamping(rpVector3D &damp)
{
	if (isFine()) 
		NewtonBodyGetAngularDamping(nBody, (f32 *)&damp);
}
