/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#include "include\rpscenemanager.h"
#include "rpGeometryBuffer.h"
#include "rpGeometryNode.h"
#include "rpLightNode.h"
#include "rpCameraNode.h"

rpSceneManager::rpSceneManager(rpVideoDriver *vD)
{
	video_driver = vD;

	resource_manager = new rpResourceManager(video_driver);

	// create the root node (with no parent)
	root_node = new rpSceneNode(NULL, this);

	active_camera = NULL;
	physics_world = NULL;

}

rpSceneManager::~rpSceneManager(void)
{
	// delete the whole hierarchy

}
// RESOURCES METHODS
rpGeometryBuffer *rpSceneManager::getGeometryBuffer(const c8* name)
{
	if (resource_manager)
		return resource_manager->getGeometryBuffer(name);
	else {
		#ifdef 	_DEBUG
		printf("rpSceneManager::getGeometryBuffer() failed! no resource_manager\n");
		#endif
		return NULL;
	}

}

// CREATION METHODS
rpGeometryNode *rpSceneManager::newGeometryNode(rpSceneNode *_parent, const c8* _file, const c8* _texture_file, bool hbuffer)
{
	// TODO: resources

	if (_parent == NULL)
		_parent = root_node;
	rpGeometryNode *gN = new rpGeometryNode(_parent, this);
	if (gN) {
		rpGeometryBuffer *gB = resource_manager->getGeometryBuffer(_file, hbuffer);
				
		// load texture if specified
		if (_texture_file)
			gN->setTexture(resource_manager->getTexture(_texture_file));

		gN->geometry_list.push_back(gB);
	}
	

	return gN;
}

rpLightNode *rpSceneManager::newLightNode(rpSceneNode *_parent)
{
	if (_parent == NULL)
		_parent = root_node;

	rpLightNode *lN = new rpLightNode(_parent, this);

	return lN;
}

rpCameraNode *rpSceneManager::newCameraNode(rpSceneNode *_parent)
{
	 rpCameraNode *cN = new rpCameraNode(_parent, this);
	 if (!active_camera)
		 active_camera = cN;

	 return cN;

}

rpSceneNode* rpSceneManager::newSceneNode(rpSceneNode *_parent)
{
	if (_parent == NULL)
		_parent = root_node;
	return new rpSceneNode(_parent, this);


}



// physics creation
rpPhysicsWorld*	rpSceneManager::newPhysicsWorld()
{
	physics_world = new rpPhysicsWorld;
	return physics_world;
}
rpPhysicsCollision*	rpSceneManager::newPhysicsCollision(E_PCOLLISION_TYPE pT, f32 sX,f32 sY ,f32 sZ,rpMatrix* offset)
{
	if (physics_world) {
		rpPhysicsCollision *pC = new rpPhysicsCollision(physics_world, pT, sX, sY, sZ, offset);

		return pC;
	}
	else {
		#ifdef 	_DEBUG
		printf("rpSceneManager::newPhysicsCollision() failed! no physics_world\n");
		#endif
		return NULL;
	}
}
rpPhysicsCollision*	rpSceneManager::newPhysicsCollision(E_PCOLLISION_TYPE pT, f32 R,f32 H,rpMatrix* offset)
{
	if (physics_world) {
		rpPhysicsCollision *pC = new rpPhysicsCollision(physics_world, pT, R, H,offset);

		return pC;
	}
	else {
		#ifdef 	_DEBUG
		printf("rpSceneManager::newPhysicsCollision() failed! no physics_world\n");
		#endif
		return NULL;
	}
}
rpPhysicsCollision*	rpSceneManager::newPhysicsCollision(E_PCOLLISION_TYPE pT, rpGeometryBuffer *mesh, rpMatrix* offset)
{
	if (physics_world) {
		rpPhysicsCollision *pC = new rpPhysicsCollision(physics_world, pT, mesh,offset);

		return pC;
	}
	else {
		#ifdef 	_DEBUG
		printf("rpSceneManager::newPhysicsCollision() failed! no physics_world\n");
		#endif
		return NULL;
	}

}

rpPhysicsNode*	rpSceneManager::newPhysicsNode(rpPhysicsCollision* pC, rpSceneNode *sN)
{
	if (physics_world && pC && pC->isFine()) {
		
		rpPhysicsNode *pN = new rpPhysicsNode(physics_world, pC);
		if (pN)
			pN->setSceneNode(sN);
		return pN;
	}
	else {
		#ifdef 	_DEBUG
		printf("rpSceneManager::newPhysicsNode() failed!\n");
		#endif
		return NULL;
	}

}

void rpSceneManager::updatePhysics()
{
	if (physics_world)
		physics_world->update();
}
void rpSceneManager::updatePhysics(f32 step)
{
	if (physics_world)
		physics_world->update(step);
}

// RENDER METHODS
bool rpSceneManager::beginScene(bool backBuffer, bool zBuffer, rpColorf color)
{
	return video_driver->beginScene(backBuffer, zBuffer, color);

}

bool rpSceneManager::endScene()
{
	return video_driver->endScene();
}

void rpSceneManager::drawAll()
{
	// go through the scene node hierarchy
	if (root_node && video_driver) {
		
		// clear all containers
		pre_geometry_list.clear();
		default_list.clear();
		
		// get rid of all dynamic lights
		// they will be activated again in the first pass
		video_driver->removeDynamicLights();
        
		// let all the nodes prepare for rendering
		root_node->preRender();

		// render current camera
		if (active_camera)
			video_driver->setTransform(E_TS_VIEW, active_camera->transformation);

		// render pre-geometry stuff like lights
		for (pre_geometry_list_it = pre_geometry_list.begin();
			 pre_geometry_list_it!= pre_geometry_list.end();
			 pre_geometry_list_it++) {
				
                 if (*pre_geometry_list_it) (*pre_geometry_list_it)->Render();
			 }
		
		
		// render geometry
		for (default_list_it = default_list.begin();
			 default_list_it!= default_list.end();
			 default_list_it++) {
				
				if ((*default_list_it).node){
					// apply texture
					if ((*default_list_it).key)				
						video_driver->setTexture((*default_list_it).key);
					// render target
					((*default_list_it).node)->Render();
				}
			 }
        
	}

}

void rpSceneManager::registerNode(const E_RENDERPASS_TYPE rpt, rpSceneNode *node, void *key)
{
	switch (rpt) {
		case E_RPT_PREGEOMETRY:
			pre_geometry_list.push_back(node);
			break;

		case E_RPT_GEOMETRY:
		{

			sDefContainer container((rpTexture *)key, node);
			default_list.push_back(container);
		}	break;
		default:
			break;
	}

}
