/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#include "include\rptextureopengl.h"

//! constructor
rpTextureOpenGL::rpTextureOpenGL(rpImage* image, bool generateMipLevels)
: Pitch(0), ImageSize(0,0), hasMipMaps(generateMipLevels),
ImageData(0), TextureName(0)
{
	#ifdef _DEBUG
    setDebugName("rpTextureOpenGL");
	#endif

	#ifdef LINUX
		hasMipMaps = false;
	#endif

	if (image)
	{
		getImageData(image);

		if (ImageData)
		{
			glGenTextures(1, &TextureName);
			copyTexture();
		}		
	}
}


//! destructor
rpTextureOpenGL::~rpTextureOpenGL()
{
	delete [] ImageData;
	glDeleteTextures(1, &TextureName);
}
//! get from an rpImage object
void rpTextureOpenGL::getFromImage(rpImage *image)
{
	if (image)
	{
		getImageData(image);

		if (ImageData)
		{
			glGenTextures(1, &TextureName);
			copyTexture();
		}		
	}
}


void rpTextureOpenGL::getImageData(rpImage* image)
{
	ImageSize = image->getDimension();
	OriginalSize = ImageSize;

	dimension2d<i32> nImageSize;
	nImageSize.Width = getTextureSizeFromSurfaceSize(ImageSize.Width);
	nImageSize.Height = getTextureSizeFromSurfaceSize(ImageSize.Height);

	if (!nImageSize.Width || !nImageSize.Height ||
		!ImageSize.Width || !ImageSize.Height)
	{
		// Could not create OpenGL Texture
		return;
	}

	ImageData = new i32[nImageSize.Width * nImageSize.Height];

	if (nImageSize == ImageSize)
	{
		if (image->getColorFormat() == E_CF_A8R8G8B8)
		{
			i32 s = nImageSize.Width * nImageSize.Height;
			i32 *t = (i32*)image->lock();
			for (i32 i=0; i<s; ++i)
				ImageData[i] = t[i];
			image->unlock();
		}
		else
		{
			// slow converting
			for (i32 x=0; x<ImageSize.Width; ++x)
				for (i32 y=0; y<ImageSize.Height; ++y)
					ImageData[y*nImageSize.Width + x] = image->getPixel(x,y).color;
		}
	}
	else
	{
		// scale texture

		i32* source = (i32*)image->lock();

		f32 sourceXStep = (f32)ImageSize.Width / (f32)nImageSize.Width;
		f32 sourceYStep = (f32)ImageSize.Height / (f32)nImageSize.Height;
		f32 sy;

		if (image->getColorFormat() == E_CF_A8R8G8B8)
		{
			// copy texture scaling
			for (i32 x=0; x<nImageSize.Width; ++x)
			{
				sy = 0.0f;

				for (i32 y=0; y<nImageSize.Height; ++y)
				{
					ImageData[(i32)(y*nImageSize.Width + x)] = source[(i32)(((i32)sy)*ImageSize.Width + x*sourceXStep)];
					sy+=sourceYStep;
				}
			}
		}
		else
		{
			// convert texture scaling, slow
			for (i32 x=0; x<nImageSize.Width; ++x)
			{
				sy = 0.0f;

				for (i32 y=0; y<nImageSize.Height; ++y)
				{
					ImageData[(i32)(y*nImageSize.Width + x)] = 
						image->getPixel((i32)(x*sourceXStep), (i32)sy).color;

					sy+=sourceYStep;
				}
			}
		}
	}		

	ImageSize = nImageSize;
}



inline bool rpTextureOpenGL::testError()
{
	#ifdef _DEBUG
	GLenum g = glGetError();
	switch(g)
	{
	case GL_NO_ERROR:
		return false;
	case GL_INVALID_ENUM:
		break;
	case GL_INVALID_VALUE:
		break;
	case GL_INVALID_OPERATION:
		break;
	case GL_STACK_OVERFLOW:
		break;
	case GL_STACK_UNDERFLOW:
		break;
	case GL_OUT_OF_MEMORY:
		break;
	};

	return true;
#endif
	return false;	
}


// copies the the texture into an open gl texture.
void rpTextureOpenGL::copyTexture()
{
	glBindTexture(GL_TEXTURE_2D, TextureName);
	if (testError())
		// Could not bind Texture

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ImageSize.Width, 
		ImageSize.Height, 0, GL_BGRA_EXT , GL_UNSIGNED_BYTE, ImageData);
	if (testError())
		// Could not glTexImage2D

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	if (hasMipMaps)
	{
		i32 ret = 0;
		
		// TODO: make this work in linux too
		#ifdef _WIN32
		ret = gluBuild2DMipmaps(GL_TEXTURE_2D, 4, ImageSize.Width, ImageSize.Height,
 		            GL_BGRA_EXT, GL_UNSIGNED_BYTE, ImageData);
		#endif

		if (ret)
		{
			#ifdef _WIN32
			// could not create OpenGl mipmaps
				
			#else
			// Could not create OpenGL texture mip maps
			#endif
		}
		else
		{
			glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
			glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		}
	}	
}


// returns the size of a texture which would be the optimize size for rendering it
inline i32 rpTextureOpenGL::getTextureSizeFromSurfaceSize(i32 size)
{
	i32 ts = 0x01;
	while(ts < size)
		ts <<= 1;

	return ts;
}


// lock function
void* rpTextureOpenGL::lock()
{
	return ImageData;
}



// unlock function
void rpTextureOpenGL::unlock()
{
	copyTexture();
}


// Returns original size of the texture.
const dimension2d<i32>& rpTextureOpenGL::getOriginalSize()
{
	return OriginalSize;
}


// Returns (=size) of the texture.
const dimension2d<i32>& rpTextureOpenGL::getSize()
{
	return ImageSize;
}


// returns color format of texture
E_COLOR_FORMAT rpTextureOpenGL::getColorFormat()
{
	return E_CF_A8R8G8B8;
}



// returns pitch of texture (in bytes)
i32 rpTextureOpenGL::getPitch()
{
	return ImageSize.Width * 4;
}



// return open gl texture name
GLuint rpTextureOpenGL::getOpenGLTextureName()
{
	return TextureName;
}


// Regenerates the mip map levels of the texture. Useful after locking and 
// modifying the texture
void rpTextureOpenGL::regenerateMipMapLevels()
{
	if (hasMipMaps)
		copyTexture();
}
// apply to video driver
void rpTextureOpenGL::applyToDriver()
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, TextureName);
}