/*****************************************************************************\
 *
 * rpLib	- graphic library for rapid development of graphical applications
 *
 * Author:	Pavel Gunia (xgunia01 @ stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/


#include "include\rpvideodriveropengl.h"
#include "include\rpTextureOpenGL.h"
#include "rpGeometryBuffer.h"
#include "rpGlGeometryBuffer.h"
#include "rpCameraNode.h"
#include <string.h>


// Win32 constructor
// TODO:
#ifdef WIN32
rpVideoDriverOpenGL::rpVideoDriverOpenGL(const dimension2d<i32>& _screenSize, HWND _window, bool _fullscreen,
										 bool _stencilBuffer, bool vsync) : rpVideoDriver(_screenSize)
{



	// init the OpenGL driver
	static	PIXELFORMATDESCRIPTOR pfd =	{
		sizeof(PIXELFORMATDESCRIPTOR),	// Size Of This Pixel Format Descriptor
		1,							// Version Number
		PFD_DRAW_TO_WINDOW |		// Format Must Support Window
		PFD_SUPPORT_OPENGL |		// Format Must Support OpenGL
		PFD_DOUBLEBUFFER,			// Must Support Double Buffering
		PFD_TYPE_RGBA,				// Request An RGBA Format
		16,							// Select Our Color Depth
		0, 0, 0, 0, 0, 0,			// Color Bits Ignored
		0,							// No Alpha Buffer
		0,							// Shift Bit Ignored
		0,							// No Accumulation Buffer
		0, 0, 0, 0,					// Accumulation Bits Ignored
		16,							// 16Bit Z-Buffer (Depth Buffer)
		StencilBuffer ? 1 : 0,		// Stencil Buffer
		0,							// No Auxiliary Buffer
		PFD_MAIN_PLANE,				// Main Drawing Layer
		0,							// Reserved
		0, 0, 0						// Layer Masks Ignored
	};

	for (int i=0; i<3; ++i)
	{
		if (i == 1)
		{
			// cannot create stencil buffer
			StencilBuffer = false;
			pfd.cStencilBits = 0;
		}
		else
		if (i == 2)
		{
			// cannot create device context
			// TODO: quit
		}

		// get hdc
		if (!(HDc=GetDC(_window)))
		{
			// cannot create GL device context
			continue;
		}

		GLuint PixelFormat;

		// choose pixelformat
		if (!(PixelFormat = ChoosePixelFormat(HDc, &pfd)))
		{
			// cannot find suitable pixel format
			continue;
		}

		// set pixel format
		if(!SetPixelFormat(HDc, PixelFormat, &pfd))
		{
			// cannot set the pixel format
			continue;
		}

		// create rendering context
		if (!(HRc=wglCreateContext(HDc)))
		{
			// cannot create GL rendering context
			continue;
		}

		// activate rendering context
		if(!wglMakeCurrent(HDc, HRc))
		{
			// cannot activate current GL rendering context
			continue;
		}

		break;
	}
	last_texture = NULL;

	// load extensions
	loadExtensions();

	glViewport(0, 0, _screenSize.Width, _screenSize.Height); // Reset The Current Viewport
	glShadeModel(GL_SMOOTH);
	
	glEnable( GL_DEPTH_TEST);
	glClearDepth(1.0f);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glDepthFunc(GL_LEQUAL);

	// set the default perspective
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, 4.0 / 3.0, 0.1f, 500.0f);

	glEnable(GL_LIGHTING);
	removeDynamicLights();
	
	
	// set vsync
	if (wglSwapIntervalEXT)
		wglSwapIntervalEXT(vsync ? 1 : 0);

	
	
}

rpVideoDriverOpenGL::~rpVideoDriverOpenGL(void)
{
}
#endif //Win32 constructor

// TODO: LINUX constructor


// methods

//! presents the rendered scene on the screen, returns false if failed
bool rpVideoDriverOpenGL::endScene()
{
	rpVideoDriver::endScene();

#ifdef WIN32
	return SwapBuffers(HDc) == TRUE;
#endif

}



//! clears the zbuffer
bool rpVideoDriverOpenGL::beginScene(bool backBuffer, bool zBuffer, rpColorf color)
{
	rpVideoDriver::beginScene(backBuffer, zBuffer, color);

	GLbitfield mask = 0;

	if (backBuffer)
	{
		glClearColor(color.r, color.g,
				color.b, color.a);

		mask |= GL_COLOR_BUFFER_BIT;
	}

	if (zBuffer)
	{
		glDepthMask(GL_TRUE);
		mask |= GL_DEPTH_BUFFER_BIT;
	}

	glClear(mask);
	return true;
}

//! queries the features of the driver, returns true if feature is available
bool rpVideoDriverOpenGL::queryFeature(E_VIDEO_DRIVER_FEATURE feature)
{
	switch (feature)
	{
	case E_VDF_BILINEAR_FILER:
		return true;
	case E_VDF_RENDER_TO_TARGET:
		return false;
	case E_VDF_HARDWARE_TL:
		return false;
	case E_VDF_MIP_MAP:
		return true;
	case E_VDF_STENCIL_BUFFER:
		return StencilBuffer;
	case E_VDF_ARB_VERTEX_PROGRAM_1:
		return ARBVertexProgramExtension;
	case E_VDF_ARB_FRAGMENT_PROGRAM_1:
		return ARBFragmentProgramExtension;
	};

	return false;
}



//! sets transformation
void rpVideoDriverOpenGL::setTransform(E_TRANSFORMATION_STATE state, rpMatrix& mat)
{
	Transformation3DChanged = true;
    
	mat.copyTo(matrix[state]);

	switch(state)
	{
	case E_TS_VIEW: {
		// OpenGL only has a model matrix, view and world is not existent. so lets fake these two.
		glMatrixMode(GL_MODELVIEW);
		rpMatrix M(matrix[E_TS_WORLD]);
		M.mult(matrix[E_TS_VIEW]);
		glLoadMatrixf((GLfloat*)&M);
		}
		break;
	case E_TS_WORLD: {
		glMatrixMode(GL_MODELVIEW);
		rpMatrix M(matrix[E_TS_WORLD]);
		M.mult(matrix[E_TS_VIEW]);
		glLoadMatrixf((GLfloat*)&M);
		}
		break;
	case E_TS_PROJECTION: 
		{
			// flip z to compensate OpenGLs right-hand coordinate system
			//mat.m12*= -1.0f;

			glMatrixMode(GL_PROJECTION);
			glLoadMatrixf((GLfloat*)&mat);
			
		}
		break;
	}
}

rpMatrix rpVideoDriverOpenGL::getTransform(E_TRANSFORMATION_STATE state)
{

	return matrix[state];
}





void rpVideoDriverOpenGL::loadExtensions()
{
	if (atof((c8*)glGetString(GL_VERSION)) >= 1.2)
	{
		// OpenGl v1.2 or better
	}
		

	const GLubyte* t = glGetString(GL_EXTENSIONS);
	i32 len = strlen((const char*)t);
	c8 *str = new c8[len+1];
	c8* p = str;

	for (i32 i=0; i<len; ++i)
	{
		str[i] = (char)t[i];

		if (str[i] == ' ')
		{
			str[i] = 0;
			if (strstr(p, "GL_ARB_multitexture"))
				MultiTextureExtension = true;
			else
			if (strstr(p, "GL_ARB_vertex_program"))
				ARBVertexProgramExtension = true;
			else
			if (strstr(p, "GL_ARB_fragment_program"))
				ARBFragmentProgramExtension = true;

			p = p + strlen(p) + 1;
		}
	}

	delete [] str;

	if (MultiTextureExtension)
	{
		#ifdef _WIN32

		// Windows
		// get multitexturing function pointers

		pGlActiveTextureARB   = (PFNGLACTIVETEXTUREARBPROC) wglGetProcAddress("glActiveTextureARB");
		pGlClientActiveTextureARB= (PFNGLCLIENTACTIVETEXTUREARBPROC) wglGetProcAddress("glClientActiveTextureARB");
		if (!pGlActiveTextureARB || !pGlClientActiveTextureARB)
		{
			MultiTextureExtension = false;
			// failed to load GL multitexturing extension
		}

		// get fragment and vertex program function pointers
		pGlGenProgramsARB = (PFNGLGENPROGRAMSARBPROC) wglGetProcAddress("glGenProgramsARB");
		pGlBindProgramARB = (PFNGLBINDPROGRAMARBPROC) wglGetProcAddress("glBindProgramARB");
		pGlProgramStringARB = (PFNGLPROGRAMSTRINGARBPROC) wglGetProcAddress("glProgramStringARB");
		pglDeleteProgramsARB = (PFNGLDELETEPROGRAMSNVPROC) wglGetProcAddress("glDeleteProgramsARB");
		pglProgramLocalParameter4fvARB = (PFNGLPROGRAMLOCALPARAMETER4FVARBPROC) wglGetProcAddress("glProgramLocalParameter4fvARB");

		#endif

		glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &MaxTextureUnits);
	}

	if (MaxTextureUnits < 2)
	{
		MultiTextureExtension = false;
		// GL device has only one texturing unit
	}

	// get vsync extension
	#ifdef _WIN32
     wglSwapIntervalEXT = (PFNWGLSWAPINTERVALFARPROC)wglGetProcAddress( "wglSwapIntervalEXT" );
	#endif
}

// device dependent texture handling
rpTexture*	rpVideoDriverOpenGL::newTexture()
{
	return new rpTextureOpenGL(NULL, true);
}

// device dependent vertex buffers
rpGeometryBuffer*	rpVideoDriverOpenGL::newGeometryBuffer()
{
	return new rpGlGeometryBuffer(0, 0, E_GEOBUFFER_OPENGL, this);
}

// render device dependent geometry buffer
void rpVideoDriverOpenGL::renderGeometryBuffer(rpGeometryBuffer *gB)
{
	
	if (gB) {

		if (gB->type == E_GEOBUFFER_OPENGL) {
			rpGlGeometryBuffer *GlGB = (rpGlGeometryBuffer *)gB;
		
			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY );
			glEnableClientState(GL_NORMAL_ARRAY );

			
			if (!(gB->hardware_buffer)) {
				// draw everything the usual way (no hardware buffers)
				switch (gB->type) {
					case E_GB_INDEXED: {
						glNormalPointer(GL_FLOAT, sizeof(sVertex3f),		(f32*)GlGB->normals.pointer());
						glTexCoordPointer(2, GL_FLOAT, sizeof(sVertex2f),	(f32*)GlGB->texcoords.pointer());
						glVertexPointer(3, GL_FLOAT, sizeof(sVertex3f),		(f32*)GlGB->vertices.pointer());

						glDrawElements(GL_TRIANGLES, GlGB->indices.size(), GL_UNSIGNED_SHORT, GlGB->indices.pointer());
						break;
					}
					case E_GB_PLAIN: {
						glNormalPointer(GL_FLOAT, sizeof(sVertex3f),		(f32*)GlGB->normals.pointer());
						glTexCoordPointer(2, GL_FLOAT, sizeof(sVertex2f),	(f32*)GlGB->texcoords.pointer());
						glVertexPointer(3, GL_FLOAT, sizeof(sVertex3f),		(f32*)GlGB->vertices.pointer());
					
						glDrawArrays(GL_TRIANGLES, 0, GlGB->vertices.size());
						break;
					}
				}
				glFlush();
			}
			else {
				// draw a hardware buffer
				
				glCallList(GlGB->buffer_ID);
				glFlush();

			}


			glDisableClientState(GL_VERTEX_ARRAY);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY );
			glDisableClientState(GL_NORMAL_ARRAY );
		}
	}

}

// handle lights
void	rpVideoDriverOpenGL::addDynamicLight(rpLightNode *light)
{
	if (!(last_set_light < GL_MAX_LIGHTS))
		return;

	if (!light)
		return;

	setTransform(E_TS_WORLD, rpMatrix());

	i32 lidx = GL_LIGHT0 + last_set_light;
	last_set_light++;
	GLfloat data[4];

	// set position
	data[0] = light->absolute_transformation.getPosX();
    data[1] = light->absolute_transformation.getPosY();
    data[2] = light->absolute_transformation.getPosZ();
	data[3] = 1.0f;
    glLightfv(lidx, GL_POSITION, data);

	// set diffuse color
	data[0] = light->diffuseColor.r;
	data[1] = light->diffuseColor.g;
	data[2] = light->diffuseColor.b;
	data[3] = light->diffuseColor.a;
	glLightfv(lidx, GL_DIFFUSE, data);

	// set specular color
	data[0] = light->specularColor.r;
	data[1] = light->specularColor.g;
	data[2] = light->specularColor.b;
	data[3] = light->specularColor.a;
	glLightfv(lidx, GL_SPECULAR, data);

	// set ambient color
	data[0] = light->ambientColor.r;
	data[1] = light->ambientColor.g;
	data[2] = light->ambientColor.b;
	data[3] = light->ambientColor.a;
	glLightfv(lidx, GL_AMBIENT, data);

	// set attenuation
	//glLightf(lidx, GL_CONSTANT_ATTENUATION, 0.0f);
	//glLightf(lidx, GL_LINEAR_ATTENUATION, 1.0f / light->radius);
	glLightf(lidx, GL_QUADRATIC_ATTENUATION, 1.0f / light->radius);

	glEnable(lidx);


}
void	rpVideoDriverOpenGL::removeDynamicLights()
{
	for (i32 i=0; i<last_set_light; i++)
		glDisable(GL_LIGHT0 + i);
	
	last_set_light = 0;

	
}

void rpVideoDriverOpenGL::setTexture(rpTexture *tex)
{
	if (last_texture) {
		if (tex)
			if (last_texture != tex) {
				tex->applyToDriver();
				last_texture = tex;
			}
	}
	else 
		if (tex) {
			tex->applyToDriver();
			last_texture = tex;
		}
}
